
// BuProlog.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020

/*  The top-level of the compiler is located in the 
    main() function at the bottom of this file. 

    Most of the code here is related to parsing -- converting the
    text of a user's program into a syntax tree, by implementing
    a recursive descent parser for Prolog based on the grammar 
    given below.
*/

/* Derived from the Prolog system by 
   Stefan Buettcher, stefan@buettcher.org, Feb. 2002,
   available at http://stefan.buettcher.org/cs/wam/.

   Buettcher's run-time implements a simplified version of
   Warren's Abstract Machine (the WAM, 1983), but supports a lot
   of Prolog built-ins,

   For details on the WAM, see 
   "Warren's Abstract Machine: A Tutorial Reconstruction" 
   by Hassan Ait-Kaci, available at 
   https://github.com/a-yiorgos/wambook

   Main changes:
     * use of a standard recursive-descent parser  -- see here
     * a greatly simplified runtime system which only
       supports a basic version of Prolog    -- see BuWAM.java
     * debugging and graphing options

   Usage:
      > java -jar BuProlog.jar -d -p lists.pro
      > java -jar BuWAM.jar -d -p lists.wam
      
      ?- member(j, [g,h,j,k]).
      ?- member(X, [j,i,m]).
      ?- q
*/

/*  Prolog EBNF grammar:

    program ::=  query | clauses
  
    query ::= '?-' body '.'
      
    clauses ::= clause clause*
    clause ::= functor [  ':-' body  ] '.'
    
    functor ::= Atom [ '(' args ')' ]

    body ::= goal [ ',' body ]
    goal ::= functor  |  '!'
    
    structure ::= Atom [ '(' args ')' ]
       // same as functor but useful separation for code gen

    args  ::=   term  ( ',' args )                        
    term  ::=   structure | Variable | Integer | list | Str

    list  ::=   '[' [  args  [ '|' list_tail ] ]  ']'     
    list_tail  ::=    Variable | list                     


    // managed by Java lex:
    Atom 	   = [a-z]([:letter:] | [:digit:])*
    Variable = [A-Z_]([:letter:] | [:digit:])*
    Integer  = [:digit:]+
    Str      = [\'\"][^\'\"]*[\'\"]

    -----------------------------------------

    For comparisons, instead of: ==  <  <=  >  >=  !=
    use: eq, lt, le, gt, ge, neq

    is/3 is written as is(v3, v1, op, v2)
    For op use:   add, sub/minus, mult/mul, div, mod
*/

import java.io.*;
import java.util.*;


public class BuProlog
{
  private BufferedReader br;
  private Lexer lexer;   // lexical analyzer
  private Token tok;        // current token

  private boolean debugOn = false; 
  private boolean queryParsing = false;


  public BuProlog(String fnm, boolean debugOn) throws IOException
  // being used to compile a program file
  { 
    this.debugOn = debugOn;

    if (fnm.endsWith(".pro")) {
      if (debugOn)
        ewriteln("\nCode Parsing:\n");
      br = new BufferedReader(new FileReader(fnm));
      lexer = new Lexer(br, true); // always show lex output
    }
    else {
      if (debugOn)
        ewriteln("Query/Term Parsing:\n");
      queryParsing = true;
      br = new BufferedReader(new StringReader(fnm)); 
                  // assume the input is a query string
      lexer = new Lexer(br, debugOn);
    }
    getToken();
  }



  // -------------------------------------------

  public ParseNode program() throws IOException
  // EBNF: program ::=  clauses
  { 
    int numSps = 1;
    if (debugOn)
      iewriteln(numSps,"program");

    ParseNode tree = new ParseNode();

    clauses(tree, numSps+1);
    if (tok.symbol() == Toks.EOF)  {
      writeln("\n\nFinished parsing."); 
      updateNames(tree);
    }
    else {
      writeln("Parsing failed");
      tree = null;
    }
    br.close();

    return tree;
  }  // end of program()



  public ParseNode query() throws IOException
  // EBNF:  query ::=  body '.'
  {
    int numSps = 1;
    if (debugOn)
      iewriteln(numSps, "query");

    ParseNode tree = new ParseNode();
    tree.type = ParseType.QUERY;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();

    body(tree.tail, numSps+1); 
    match(tok, Toks.PERIOD);
    if (tok.symbol() == Toks.EOF)  {
      // writeln("\nFinished parsing."); 

      tree.head.type = ParseType.PREDICATE;
      tree.head.tail = null;
      tree.head.head = new ParseNode();
      tree.head.head.type = ParseType.ATOM;
      tree.head.head.value = "qqquery/0_1/1";  // /arity_count/total
      tree.head.tail = null;
    }
    else {
      writeln("Parsing failed");
      tree = null;
    }
    br.close();

    return tree;
  }  // end of query()



  private void clauses(ParseNode tree, int numSps) throws IOException
  // EBNF:  clauses ::= clause clause*
  {
    if (debugOn)
      iewriteln(numSps, "clauses");

    tree.type = ParseType.CLAUSES;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();

    clause(tree.head, numSps+1);

    ParseNode temp = tree;
    while (tok.symbol() != Toks.EOF) {
      temp = temp.tail;
      temp.type = ParseType.CLAUSES;
      temp.head = new ParseNode();
      temp.tail = new ParseNode();
      clause(temp.head, numSps+1); 
    }
    temp.tail = null;
  }  // end of clauses()



  private void clause(ParseNode tree, int numSps) throws IOException
  // EBNF:  functor [  ':-' body  ] '.'
  {
    if (debugOn)
      iewriteln(numSps, "clause");

    tree.type = ParseType.CLAUSE;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();

    functor(ParseType.PREDICATE, tree.head, numSps+1);
    if (tok.symbol() == Toks.IF) {
      match(tok, Toks.IF);
      body(tree.tail, numSps+1); 
    }
    else
      tree.tail = null;
    match(tok, Toks.PERIOD);
  }  // end of clause()



  private void functor(ParseType nodeType, ParseNode tree, int numSps) throws IOException
  // EBNF:  functor ::= Atom [ '(' args ')' ]
  /* nodeType is PREDICATE (for head) 
     or CALL (for body goal)
  */
  {
    if (debugOn)
      iewriteln(numSps, "functor");

    tree.type = nodeType;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();
  
    atom(tree.head, numSps+1);
    if (tok.symbol() == Toks.LPAREN) {
      match(tok, Toks.LPAREN);
      int numArgs = arguments(tree.tail, false, numSps+1);   // not in a list 
      if (!Builtin.contains(tree.head.value))  // user-defined pred
        tree.head.value = tree.head.value + "/" + numArgs;
      match(tok, Toks.RPAREN);
    }
    else {
      if (!Builtin.contains(tree.head.value))  // user-defined pred
        tree.head.value = tree.head.value + "/0";
      tree.tail = null;
    }
  }  // end of functor()



  private void body(ParseNode tree, int numSps) throws IOException
  // EBNF:  body ::= goal [ ',' body ]
  {
    if (debugOn)
      iewriteln(numSps, "body");

    tree.type = ParseType.BODY;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();

    goal(tree.head, numSps+1);
    if (tok.symbol() == Toks.COMMA) {
      match(tok, Toks.COMMA);
      body(tree.tail, numSps+1); 
    }
    else
      tree.tail = null;
  }  // end of body()


  private void goal(ParseNode tree, int numSps) throws IOException
  // EBNF:  goal ::= functor   |  '!'
  {
    if (debugOn)
      iewriteln(numSps, "goal");

    if (tok.symbol() == Toks.ATOM)
      functor(ParseType.CALL, tree, numSps+1);
    else if (tok.symbol() == Toks.CUT) {
      tree.type = ParseType.CALL;
      tree.head = new ParseNode();
      tree.head.value = "CUT";
      match(tok, Toks.CUT);
    }
    else
      error(lexer.position(), "goal expected");
  }  // end of goal()




  private int arguments(ParseNode tree, boolean inList, int numSps) 
                                            throws IOException
  // EBNF: args  ::=   term  ( ',' args )
  {
    if (debugOn)
      iewriteln(numSps, "arguments");

    if (inList)
      tree.type = ParseType.LIST_ARGS;
    else
      tree.type = ParseType.ARGS;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();

    int numArgs = 1;
    term(tree.head, numSps+1);
    if (tok.symbol() == Toks.COMMA) {
      match(tok, Toks.COMMA);
      numArgs += arguments(tree.tail, inList, numSps+1);
    }
    else
      tree.tail = null;
    return numArgs;
  }  // end of arguments()




  public void term(ParseNode tree, int numSps) throws IOException
  /* EBNF:   term  ::=   structure | Variable | Integer | list | Str

     public so can be used by BuWAM's readTerm()
  */
  {
    if (debugOn)
      iewriteln(numSps, "term");

    if (tok.symbol() == Toks.ATOM) {
      structure(tree, numSps+1);
    }
    else if (tok.symbol() == Toks.VAR) {
      var(tree, numSps+1);
    }
    else if (tok.symbol() == Toks.INTEGER) {
      integer(tree, numSps+1);
    }
    else if (tok.symbol() == Toks.LSQUARE) {
      list(tree, numSps+1);
    }
    else if (tok.symbol() == Toks.STR) {
      str(tree, numSps+1);
    }
    else
      error(lexer.position(), "term expected");
  }  // end of term()





  private void structure(ParseNode tree, int numSps) throws IOException
  /* EBNF: structure ::= Atom [ '(' args ')' ]
  */
  {
    if (debugOn)
      iewriteln(numSps, "structure");

    tree.type = ParseType.STRUCTURE;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();
  
    atom(tree.head, numSps+1);
    if (tok.symbol() == Toks.LPAREN) {
      tree.head.type = ParseType.CONSTANT;  // was ATOM
      match(tok, Toks.LPAREN);
      arguments(tree.tail, false, numSps+1);   // not in a list
      match(tok, Toks.RPAREN);
    }
    else {   // replace this node with a constant
      tree.type = ParseType.CONSTANT;
      tree.value = tree.head.value;
      tree.head = null;
      tree.tail = null;
    }
  }  // end of structure()



  private void list(ParseNode tree, int numSps) throws IOException
  /* EBNF: list  ::=   '[' [  args  [ '|' list_tail ] ]  ']'
  */
  {
    if (debugOn)
      iewriteln(numSps, "list");

    tree.type = ParseType.LIST;
    tree.head = new ParseNode();
    tree.tail = new ParseNode();
  
    match(tok, Toks.LSQUARE);
    if (tok.symbol() != Toks.RSQUARE) {
      arguments(tree.head, true, numSps+1);   //  list arguments
      if (tok.symbol() == Toks.BAR) {
        match(tok, Toks.BAR);
        listTail(tree.tail, numSps+1);
      }
      else 
        tree.tail = null;
      match(tok, Toks.RSQUARE);
    }
    else {   // an empty list
      match(tok, Toks.RSQUARE);
      tree.value = "[]";
      tree.head = null;
      tree.tail = null;
    }
  }  // end of list()




  private void listTail(ParseNode tree, int numSps) throws IOException
  /* EBNF: list_tail  ::=  Variable | list
  */
  {
    if (debugOn)
      iewriteln(numSps, "listTail");

    if (tok.symbol() == Toks.VAR) {
      var(tree, numSps+1);
    }
    else if (tok.symbol() == Toks.LSQUARE) {
      list(tree, numSps+1);
    }
    else
      error(lexer.position(), "list tail expected");
  }  // end of listTail()




  private void atom(ParseNode tree, int numSps) throws IOException
  {
    if (debugOn)
      iewriteln(numSps, "atom");

    if (tok.symbol() == Toks.ATOM) {
      String atom = tok.lexeme();
      tree.type = ParseType.ATOM;
      tree.value = atom;
      getToken();
    }
    else
      error(lexer.position(), Toks.ATOM.getText() + " expected");
  }  // end of atom()



  private void var(ParseNode tree, int numSps) throws IOException
  {
    if (debugOn)
      iewriteln(numSps, "var");

    if (tok.symbol() == Toks.VAR) {
      String var = tok.lexeme();
      tree.type = ParseType.VARIABLE;
      tree.value = var;
      getToken();
    }
    else
      error(lexer.position(), Toks.VAR.getText() + " expected");
  }  // end of var()



  private void integer(ParseNode tree, int numSps) throws IOException
  {
    if (debugOn)
      iewriteln(numSps, "integer");

    if (tok.symbol() == Toks.INTEGER) {
      String intStr = tok.lexeme();
      tree.type = ParseType.CONSTANT;
      tree.value = intStr;
      getToken();
    }
    else
      error(lexer.position(), Toks.INTEGER.getText() + " expected");
  }  // end of integer()



  private void str(ParseNode tree, int numSps) throws IOException
  {
    if (debugOn)
      iewriteln(numSps, "str");

    if (tok.symbol() == Toks.STR) {
      String s = tok.lexeme();
      tree.type = ParseType.STR;
      tree.value = s;
      getToken();
    }
    else
      error(lexer.position(), Toks.STR.getText() + " expected");
  }  // end of str()



  // ------------------- parse utils --------------------------



  private void getToken() throws IOException
  {  tok = lexer.nextToken(); }



  private void match(Token t, Toks match) throws IOException
  {
    if (tok.symbol() == match) 
      getToken(); 
    else 
      error(lexer.position(), match.getText() + " expected");
  }  // end of match()



  private void error(int pos, String msg)  throws IOException
  {
    if (!queryParsing)
      writeln("");
    for (int i = 0; i < pos; i++) 
      write(" ");
    if (queryParsing)
      write("   ");   // to compensate for "?- "
    writeln("^");
    writeln("Parsing error: " + msg);
    throw new IOException("Parsing error");
  }  // end of error()


  // ----------------------- I/O utils ----------------------


  private void write(String s)
  {  System.out.print(s);  } 


  private void writeln(String s)
  {  System.out.println(s);  } 



  private void ewrite(String s)
  {  System.err.print(s);  } 


  private void ewriteln(String s)
  {  System.err.println(s);  } 


  private void iewriteln(int numSps, String s)
  { 
    if (numSps == 0)
      System.err.printf("\t\t\t%c%s\n", '%', s);
    else
      System.err.printf("\t\t\t%c%"+numSps+"s%s\n", '%', "", s);
  }



  private void saveString(String fnm, String s) 
  {
    if (s == null) {
      writeln("No data to save in " + fnm);
      return;
    }
    try {
      FileWriter fw = new FileWriter( new File(fnm));
      fw.write(s);
      fw.close();
      writeln("Saved code to file: " + fnm);
    } 
    catch (IOException ex) {
      writeln("Could not save code to file: " + fnm);
    } 
  }  // end of saveString()



  // ------------ predicate name  modification -------------------------


  private void updateNames(ParseNode tree)
  /* add indicies to predicate names
     e.g.  member/2 becomes member/2_1/2
           where the first number is the index, and the second is
           total no. of clauses
  */
  {
    TreeMap<String,Integer> predCounts = new TreeMap<>();
            // holds <predname, no-of-clauses> pairs
    ParseNode t;

    // add indicies of first pass over tree
    if ((tree.type == ParseType.CLAUSES) && (tree.head != null)) {
      t = tree;
      do {
        ParseNode predAtom = t.head.head.head;   // clause->predicate->atom
        String pnm = predAtom.value;
        int cnt = 0;
        if (predCounts.containsKey(pnm))
          cnt = predCounts.get(pnm);
        else
          predCounts.put(pnm, 0);

        predCounts.put(pnm, ++cnt);
        predAtom.value = pnm + '_' + cnt;
        t = t.tail;
      } while (t != null);
    }

    // add clause totals on second pass
    if ((tree.type == ParseType.CLAUSES) && (tree.head != null)) {
      t = tree;
      do {
        ParseNode predAtom = t.head.head.head;    // clause->predicate->atom
        String pnm = predAtom.value;
        if (pnm.indexOf('_') > 0)
          pnm = pnm.substring(0, pnm.indexOf('_'));
        predAtom.value += "/" + predCounts.get(pnm);
        t = t.tail;
      } while (t != null);
    }
  } // end of updateNames()





  // ----------------------------------------------
  /* This compiler translates the user's Prolog program into 
     a syntax tree made up of ParseNode objects.
     
     This tree is translated into WAM code by a call to
     ParseNode.codeGen(), which recurses over the tree.  
     
     The generated WAM code is stored a a list of WStatement objects
     in a WProgram object.
  */



  public static void main(String args[])
  {
    // command line processing of arguments
    if ((args.length < 1) || (args.length > 3)) {
      System.out.println("Usage: java -jar BuProlog.jar [-d] [-p] <Prolog file>");
      return;
    }

    String fnm;   // the user's Prolog program
    boolean[] flags = new boolean[2];  // { debugOn, picOn }; both false

    if (args.length == 1)
      fnm = args[0];
    else if (args.length == 2) {
      flags = setFlags(new String[]{args[0], ""});
      fnm = args[1];
    }
    else {
      flags = setFlags(new String[]{args[0], args[1]});
      fnm = args[2];
    }
    boolean debugOn = flags[0];
    boolean picOn = flags[1]; 

    // parse the input file, generating a tree
    BuProlog p = null;
    ParseNode tree = null;
    try {
      p = new BuProlog(fnm, debugOn);
      tree = p.program();   // the parse tree
    }
    catch(IOException e) {
      // System.out.println(e);
      System.exit(1);
    }
      
    // generate a drawing of the parse tree
    if (picOn) {
      System.err.println("\nDrawing parse tree");
      CallGraph cg = new CallGraph();
      tree.draw(cg, "main");
      cg.end();
      cg.save("parseTree.png");
      System.err.println("Saved to parseTree.png\n");
    }


    // translate the tree into WAM code
    if (debugOn)
      System.err.println("\nGenerating WAM Code:\n");
    WAMProgram prog = new WAMProgram(tree, debugOn);
    if (prog == null)
      System.out.println("WAM code could not be generated");
    else {
      // save the code as text to a "wam" file
      if (debugOn) {
        System.err.println("\n\n----------------- WAM code ----------------");
        System.err.println(prog.toString());
        System.err.println("--------------------------------------------\n");
      }
      int extIdx = fnm.lastIndexOf('.');
      String name = fnm;
      if (extIdx != -1)
        name = fnm.substring(0, extIdx);
      p.saveString(name+".wam", prog.toString());
    }
  }  // end of main()



  private static boolean[] setFlags(String[] args)
  {
    boolean debugOn = false;
    boolean picOn = false;

    for (String arg : args) {
       switch(arg.toUpperCase()) {
          case "-D": case "--D":
            debugOn = true; break;

          case "-P": case "--P":
            picOn = true; break;

          case "-DP":   case "-PD":
          case "--DP":  case "--PD":
            debugOn = true; picOn = true; break;

          case "": break;

          default: 
            System.out.println("Unknown flag: " + arg);
            break;
       }
    }
    return new boolean[] { debugOn, picOn };
  }  // end of setFlags()


}  // end of BuProlog class
