
// BuWAM.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, July 2020
// July 24th: added read/1

// A run-time system based on Buettcher's WAM.

/* Derived from the Prolog system by 
   Stefan Buettcher, stefan@buettcher.org, Feb. 2002,
   available at http://stefan.buettcher.org/cs/wam/.

   For details on the WAM, see 
   "Warren's Abstract Machine: A Tutorial Reconstruction" 
   by Hassan Hvt-Kaci, available at 
   https://github.com/a-yiorgos/wambook

   Main changes:
     * use of a standard recursive-descent parser  -- see BuProlog.java
     * a greatly simplified runtime system which only
       supports a basic version of Prolog 
     * debugging and graphing options

   Usage:
      > java -jar BuProlog.jar -d -p lists.pro
      > java -jar BuWAM.jar -d -p lists.wam
      
      ?- member(j, [g,h,j,k]).
      ?- member(X, [j,i,m]).
      ?- q.

      > java -jar BuWAM.jar -d -p family.wam 2> debug.txt
      ?- grandParent(GP, jane).

*/


import java.awt.*;
import java.io.*;
import java.util.*;



public class BuWAM
{
  private static int MAX_OPS = 5000000; 
     // an operator overflow limit used in runWAM()


  private WAMCode wc;        // the loaded WAM code
  private int pc = 0;        // program counter
  private boolean queryFailed; 

  private ArrayList<Var> queryVars = new ArrayList<>();    
          // vars appearing in the query

  private ArrayList<Var> args = new ArrayList<>();  
         // vars used as arguments passed into a clause and to a call

  private ArrayList<Env> envList = new ArrayList<>();
         // each Env object stores the vars local to a clause
  private int envIdx;    // list's index to the current environment

  private Trail trail = new Trail();   
             // an undo-list of all variable bindings

  private int cp;   // continuation pointer (i.e. the return address)


  private Stack<ChoicePoint> choices = new Stack<>();
  private ChoicePoint cutPoint = null; 
                   // current choicepoint; used by the cut WAM op
  

  // -------- vars used for debugging and graphing 
 
  private boolean debugOn;     // display debug info?
  private boolean picOn;       // display graph info?


  private int nodesCount = -1;   
     // number of call and try/trust WAM ops executed

  private boolean hasDeallocated = false;  
      // used to toggle debug/graph info on deallocate and proceed WAM ops

  private CallGraph cg;   // display query execution as a graph
  private int numPicsSaved = 0;
  private String tryNodeNm, callNodeNm;


  public BuWAM(String fnm, boolean debugOn, boolean picOn)
  { 
    this.debugOn = debugOn;
    this.picOn = picOn;
    wc = new WAMCode(fnm, debugOn);
    reset();
  } // end of BuWAM()


  private void reset()
  {
    queryVars.clear();
    args.clear();
    envList.clear(); 
    envIdx = -1;   
    trail.reset();
    cp = -1; 
    choices.clear();
    cutPoint = null;

    nodesCount = -1;   // so query is node '0'
    if (picOn)
      cg = new CallGraph();
    callNodeNm = "query";
    tryNodeNm = null;
  } // end of reset()



  // ------------------  run a query -------------------


  public boolean runQuery(String queryStr)
  /* compile a query string into WAM code; 
     add it to the end of the program code
     and start its execution
  */
  {
    if (queryStr == null) {
      writeln("Query string is null");
      return true; 
    }
    queryStr = queryStr.trim();
    if (queryStr.isEmpty()) {
      writeln("Query string is empty");
      return true;
    }

    if ( queryStr.equals("quit") || queryStr.equals("q") ||
         queryStr.equals("quit.") || queryStr.equals("q.") ) 
      return false;  // stop processing

    // if we get here then the line is a new query
    if (debugOn)
      ewriteln("QUERY: " + queryStr);

    reset();

    WAMClause query = parseQuery(queryStr);
    if (query == null)
     return true;   // a parsing error so move on to next query

    executeQuery(query); 

    // reset();
    return true;
  } // end of runQuery()



  private WAMClause parseQuery(String queryStr)
  {
    WAMClause query = null;
    try {
      BuProlog qc = new BuProlog(queryStr, debugOn);
      ParseNode queryTree = qc.query();

      if (picOn) {
        // generate a drawing of the query's parse tree
        CallGraph cgPT = new CallGraph();
        queryTree.draw(cgPT, "main");
        cgPT.end();
        cgPT.save("queryParse.png");
      }
      if (debugOn)
        ewriteln("\n\nCode generation:\n");
      query = new WAMClause(queryTree, debugOn);
    }
    catch(IOException e)   // usually a parsing error
    {  return null;   }

    boolean isSaved = saveString("queryWAM.txt", 
                             "\n% ?- " + queryStr + 
                             "\n\n" + query.toString());
    if (debugOn) {
      if (isSaved)
        ewriteln("Saved query to queryWAM.txt");
      else
        ewriteln("Could not save query to queryWAM.txt");
    }
    if (picOn)
      cg.comment("query", "Q: " + queryStr, CallGraph.LIME);

    return query;
  }  // end of parseQuery()



  private void executeQuery(WAMClause query)
  {
    wc.deleteQuery();   // remove any previous query from the code

    // add the query code to the end of the program code
    wc.addQuery(query); 

    pc = wc.getQueryPos();
    String answer = "";
    do {
      runWAM();

      if (queryFailed) { 
        writeln("No.");
        if (debugOn) {
          ewriteln("--> No");
          ewriteln("due to (" + nodesCount + ")");
          ewriteln("------------------ end of query -----------------\n");
        }
        if (picOn) {
          cg.comment("query", "(" + nodesCount + ") No", CallGraph.PINK);
          cg.setColor(tryNodeNm, CallGraph.PINK);
        }
        break;
      }

      // display query variable values
      writeln(getQVars());
      if (debugOn) {
        ewrite("\n--> Result: \n" + getQVars());
        ewriteln(" due to (" + nodesCount + ")");
        printTreeVars(queryVars, queryVars.size());
        ewriteln("------------------ end of query -----------------\n");
      }
      if (picOn) {
        cg.comment("query", "(" + nodesCount + ") " + getQVars(), CallGraph.LIME);
        cg.setColor(tryNodeNm, CallGraph.LIME);
      }

      // if there are some choicepoints, 
      // ask the user if they should be tried via backtracking
      if (!choices.isEmpty()) {
        write("\nMore? (y/n) ");
        answer = readline();
        if (debugOn)
           ewriteln("More results: " + answer);
      }
      else
        break;

      // has the user decided to backtrack?
      if (answer.equals("y") ||  answer.equals("yes")) {
        if (debugOn) 
          ewriteln("  --> BACKTRACKED by USER");
        backtrack();
      }
    } while ( answer.equals("y") || answer.equals("yes"));

    if (picOn) {
      cg.end();
      numPicsSaved++;
      cg.save("queryTree" + numPicsSaved + ".png");
    }
  }  // end of executeQuery()



  private String getQVars()
  /* only print the query variables which have names;
     Format is Name=Value\n
  */
  {
    boolean hasNames = false;
    int numVars = queryVars.size();

    StringBuilder sb = new StringBuilder("");
    for (int i = 0; i < numVars; i++) {
      Var v = queryVars.get(i);
      if (v.name != null) {
        sb.append(v.toString() + "\n");   // toString() return "Name=Value"
        hasNames = true;
      }
    }
	if (!hasNames)
      sb.append("\nYes");

    // if (choices.isEmpty())
    //  sb.append("\nNo more answers");

    return sb.toString();
  }  // end of getQVars()





  public void runWAM()
  /* execute the query starting at the current pc value.
     For each WAMStatement, execute the relevant WAM 
     operation.
  */
  {
    int opCount = 0;

    while (pc >= 0) {  // pc < 0 happens on a jump error
      queryFailed = false;
      WAMStatement s = wc.getStatement(pc); 

      if (debugOn)  
        ewriteln("Run: (" + String.format("%04d", pc) + 
                    ")  " + s.toString() +
                    "  % in " + s.getClauseLabel());

      // prevent the WAM from going into an infinite execution
      if (opCount++ > MAX_OPS) {
        writeln("Maximum op count reached: " + MAX_OPS);
        queryFailed = true;
        return;
      }

      // extract the WAM operation and execute the corresponding method
      OpWam op = s.getOp();
      // PUT ops
      if (op == OpWam.PUT_VALUE) 
        put_value(s.getArg(0), s.getArg(1));
      else if (op == OpWam.PUT_CONSTANT) 
        put_constant(s.getArg(0), s.getArg(1));
      //
      else if (op == OpWam.CREATE_VARIABLE)   
           // NEW (i.e. not part of the original WAM);
           //  used to link a PERM variable to a query variable
        setQVarName(s.getArg(0), s.getArg(1));
      // GET ops
      else if (op == OpWam.GET_VARIABLE) 
        get_variable(s.getArg(0), s.getArg(1));
      else if (op == OpWam.GET_VALUE) 
        get_value(s.getArg(0), s.getArg(1));
      else if (op == OpWam.GET_CONSTANT) 
        get_constant(s.getArg(0), s.getArg(1));
      //  UNIFY ops
      else if (op == OpWam.UNIFY_VARIABLE)
        unify_variables(s.getArg(0), s.getArg(1));
      else if (op == OpWam.UNIFY_LIST)        // NEW
        unify_list(s.getArg(0), s.getArg(1), s.getArg(2));
      else if (op == OpWam.UNIFY_STRUC)       // NEW
        unify_struc(s.getArg(0), s.getArg(1), s.getArg(2));
      // clause execution ops
      else if (op == OpWam.ALLOCATE) 
        allocate();
      else if (op == OpWam.DEALLOCATE) {
        deallocate();
        hasDeallocated = true;
      }
      else if (op == OpWam.CALL) {
        nodesCount++;
        call(s.getArg(0), s.getJump());
      }
      else if (op == OpWam.PROCEED) {   // return from a clause
        pc = cp; 
        int numClauseArgs = getArity(s.getClauseLabel());
        if (debugOn) {
          ewriteln("  --> Return to " + 
                  wc.getStatement(pc).getClauseLabel());
          if (!hasDeallocated)    // not already shown return info
             printVars(args,numClauseArgs);
        }
        if (picOn) {
          if (!hasDeallocated)
            cg.comment(tryNodeNm, "ret: " + 
                     brief( getArgVals(numClauseArgs), 25));
        }
        if (hasDeallocated)
          hasDeallocated = false;
      }
      //
      else if ((op == OpWam.TRY_ME_ELSE) ||
               (op == OpWam.RETRY_ME_ELSE) ) {
        // these ops are treated identically
        nodesCount++;
        try_me_else(s.getLabel(), s.getJump());
      }
      // 
      else if (op == OpWam.TRUST_ME) {
        /* If the code following this instruction fails, 
           then fail the execution of the goal against 
           the entire predicate.
        */
        nodesCount++;
        String callLabel = s.getLabel();
        if (debugOn) {
          if (!callLabel.equals("qqquery_0"))
            ewriteln("  --> Trying " + callLabel + 
                 " (" + nodesCount + "); no choice added"); 
        }
        if (picOn)
          if (!callLabel.equals("qqquery_0"))
            tryNode(callLabel);

        pc++;
      }
      // CUT ops
      else if (op == OpWam.GET_LEVEL) 
        get_level(s.getArg(0));
      else if (op == OpWam.CUT) {
        nodesCount++;
        if (picOn)
          callNode("cut");
        cut(s.getArg(0));
      }
      //
      else if (op == OpWam.HALT)   // NEW
        break;   // the query's execution is terminated
      else if (op == OpWam.NO_OP) 
        pc++;
      //
      else { 
        writeln("Invalid operation on line " + String.format("%04d", pc));
        backtrack();
      }
    }  // end of while (pc >= 0)
    
    if (queryFailed) {
      while (!choices.isEmpty()) 
        backtrack();
    }
  } // end of runWAM()



  // -------------- WAM OPERATIONS -------


  // --------------------- Put instructions ------------------

  /* used to load values from PERM variables in a clause 
     into its ARG or QUERY variables.
  */

  private void put_value(String sv, String sAv)
  /* put_value v Av. 
     Copy the contents of v into Av
  */
  { Var v = varLookup(sv);
    Var Av = varLookup(sAv);
    Av.copyValue(v);
    pc++;
  } 


  private void put_constant(String c, String sAv)
  /* put_constant c Av
     Assign c to Av.
  */
  { Var Av = varLookup(sAv);
    Av.tag = Var.ValType.CON;
    Av.con = c;
    pc++;
  } 



  private void setQVarName(String sv, String name)
  /* NEW; not an original WAM instruction
     Assign a name string to a QUERY variable.
     This name is used when printing the variable's value.
  */
  { if (!name.equals("_")) { 
      /* do not assign the name "_" 
         so the variable's value will not be printed
      */
      Var v = varLookup(sv);
      v.name = name;
    }
    pc++;
  } // end of setQVarName()


  // --------------------- Get instructions ------------------
  /* used to load values from
     the ARG variables into PERM variables.
  */


  private void get_variable(String sv, String sAv)
  /* get_variable v Av 
     v := Av. 
     Only used for the first occurrence of v inside 
     the clause’s head.
  */
  { Var v = varLookup(sv);
    Var Av = varLookup(sAv);
    v.copyValue(Av);
    pc++;
  } 


  private void get_value(String sv, String sAv)
  /* get_value v Av 
     Does v = Av?
     Used for subsequent occurrences of Y 
     inside the clause’s head.
  */
  {  unify_variables(sv, sAv);  }



  private void get_constant(String c, String sAv)
  /*  get_constant c Av 
      Does Av == c?
  */
  { 
    Var Av = varLookup(sAv).deref();
    boolean hasFailed = true;

    if (Av.tag == Var.ValType.REF) {
      trail.add(Av);
      Av.tag = Var.ValType.CON;
      Av.con = c;
      hasFailed = false;
    }
    else if (Av.tag == Var.ValType.CON) {      
      if (c.equals(Av.con))
        hasFailed = false;
    }
    if (hasFailed)
      backtrack();
    else
      pc++;
  } // end of get_constant()



  // --------------------- Unify instructions ------------------



  private void unify_variables(String s1, String s2)
  /*  unify_variable v1 v2 
      v1 = v2 
  */
  { Var v1 = varLookup(s1);
    Var v2 = varLookup(s2);
    if (unify(v1, v2))
      pc++;
    else
      backtrack();
  } // end of unify_variables()



  private boolean unify(Var v1, Var v2)
  {
    if ((v1 == null) || (v2 == null)) 
      return false;

    v1 = v1.deref();
    v2 = v2.deref();
    if (v1 == v2) 
      return true;

    if (v1.tag == Var.ValType.REF) {
      trail.add(v1);
      v1.copyValue(v2);
      return true;
    }

    if (v2.tag == Var.ValType.REF) {
      trail.add(v2);
      v2.copyValue(v1);
      return true;
    }

    if ((v1.tag == Var.ValType.CON) && (v2.tag == Var.ValType.CON))
      return (v1.con.equals(v2.con));

    if (((v1.tag == Var.ValType.LIST) && (v2.tag == Var.ValType.LIST)) || 
        ((v1.tag == Var.ValType.STRUC) && (v2.tag == Var.ValType.STRUC)))
      if ((unify(v1.head, v2.head)) && 
         (unify(v1.tail, v2.tail)))
        return true;

    return false;
  } // end of unify()



  private void unify_list(String l, String h, String t)
  // NEW; not an original WAM instruction
  /*  unify_list list head tail 
      list = [ head | tail ] 
  */
  { Var listV = varLookup(l);
    Var head = varLookup(h);
    Var tail = varLookup(t);
    if (unify_list2(listV, head, tail))
      pc++;
    else
      backtrack();
  } 


  private boolean unify_list2(Var listV, Var head, Var tail)
  {
    if ((listV == null) || (head == null) || (tail == null)) 
      return false;

    listV = listV.deref();
    head = head.deref();
    tail = tail.deref();

    if (listV.tag == Var.ValType.REF) {
      trail.add(listV);
      listV.tag = Var.ValType.LIST;
      listV.head = head;
      listV.tail = tail;
      return true;
    }

    if (listV.tag == Var.ValType.LIST) {
      if (unify(head, listV.head))
        if (unify(tail, listV.tail))
          return true;
    }
    return false;
  } // end of unify_list2()



  private void unify_struc(String s, String sNm, String sArgs)
  // NEW; not an original WAM instruction
  /* unify_struc struct v1 v2 v3 
     v1 = v2(v3)
     v3 is a list of arguments.
  */
  { Var structV = varLookup(s);
    Var nmV = varLookup(sNm);
    Var argsV = varLookup(sArgs);
    if (unify_struc2(structV, nmV, argsV))
      pc++;
    else
      backtrack();
  }


  private boolean unify_struc2(Var structV, Var nmV, Var argsV)
  // also used to implement the univ() builtin ( =..)
  {
    if ((structV == null) || (nmV == null) || (argsV == null)) 
      return false;

    structV = structV.deref();
    nmV = nmV.deref();
    argsV = argsV.deref();
    
    if (structV.tag == Var.ValType.REF) {
      trail.add(structV);
      structV.tag = Var.ValType.STRUC;

      structV.head = new Var(nmV);
      structV.head.tag = Var.ValType.CON;
      if (!isAtom(nmV.con))
        System.out.println("Error: structure does not use an atom");
      structV.tail = new Var(argsV);
      return true;
    }

    if (structV.tag == Var.ValType.STRUC) {
      if (unify(nmV, structV.head))
        if (unify(argsV, structV.tail))
          return true;
    }
    return false;
  } // end of unify_struc2()



  private Var varLookup(String name)
  /* returns the Var associated with the supplied name;
     The format is Y or A or Q followed by a number.
  */
  {
    ArrayList<Var> vars;
    VarType vt = VarType.getType(name);
    switch (vt) {
      case PERM:
        vars = envList.get(envIdx).vars;    
                 // local vars in the clause body, which are stored
                 // in the clause's current environment
        break;
      case ARG:
        vars = args;       // vars used as arguments
        break;
      case QUERY: 
        vars = queryVars;  // vars in the query
        break;
      default: 
        writeln("Unknown type for variable " + name);
        return null;
    }

    int idx = VarType.getIndex(name);
    if (idx == -1) {
      writeln("Could not find index for variable " + name);
      return null;
    }

    // pad out the vars list until there is a var at this index pos
    int sz = vars.size();
    while (sz < idx+1) {
      vars.add(new Var(vt, sz));
      sz++;
    }

    return vars.get(idx);
  } // end of varLookup()




  // -------------- Control instructions ------------------


  private void allocate()
  /* Allocate a new environment for a clause.
  */
  { 
    if (debugOn)
      printEnvs();

    String clauseLabel = getLineLabel(pc-1);  
       // the clause's label is on the line before the "allocate" command

    Env env = new Env(clauseLabel, cp, envIdx, tryNodeNm, nodesCount);
        // tryNodeNm initialized in tryNode() to this clause
        // and assigned this nodesCount

    if ((envIdx > -1) && (!choices.isEmpty())) {
      int ce = choices.peek().envIdx;  
      if (envIdx > ce)   // delete unneeded environments
        envList.subList(envIdx+1, envList.size()).clear();
    }
    envList.add(env);
    envIdx = envList.size()-1;   // set to top of list

    pc++;
  }   // end of allocate()



  private void deallocate()
  /* Restore the previous environment, but do not delete the 
     current one. since it may 
     be needed later during backtracking.
  */
  { 
    String varBindings = getVAs(envList.get(envIdx).vars);

    if (debugOn) {
      if (envIdx != -1) {
        if (!varBindings.equals("")) {
          ewriteln("  --> ret: " + varBindings);
          ewriteln("      due to (" + nodesCount + ")");
        }
      }
      printEnvs();
      int ce = envList.get(envIdx).ce;
         // the continuation environment index (what env will be returned to)
      ewriteln("      --> leaving " + envLabel(envIdx) + 
                            " for " + envLabel(ce) );
    }
    if (picOn) {
      if (envIdx != -1) {
        if (!varBindings.equals(""))
          cg.comment(getTryNode(envIdx), 
               "ret: " + brief(varBindings, 25) + 
               "\\ndue to (" + nodesCount + ")");
      }
    }

    cp = envList.get(envIdx).cp;   
                  // the continuation pointer (the return address)
    envIdx = envList.get(envIdx).ce;  
    pc++;
  }  // end of deallocate()



  private void call(String callLabel, int pos)
  /* Call the predicate label/arity by jumping to its first clause.
     If pos is negative then it may be the 'address' of a 
     built-in predicate.
     Also used by the builtin callStruct()
  */
  { 
    if (debugOn) {
      ewriteln("  --> Calling " + getLineLabel(pos) + 
                                     " (" + nodesCount + ")");
      printVars(args, getArity(callLabel));
    }
    if (picOn)
      callNode(callLabel);

    if (pos >= 0) {  
      cp = pc+1;  // return to next line after current pc location
      if (choices.empty())
        cutPoint = null;
      else
        cutPoint = choices.peek();   // used by possible future cut 
      pc = pos;
    }
    else  // pos < 0 indicates either a built-in predicate or an error
      if (Builtin.contains(pos))
        callBuiltins(pos);
      else {
        // System.out.println("Error: Could not identify call label");
        backtrack();
      }
  } // end of call()




  // -------------- Choice instructions -----------------

  /* Try out the clauses that 
     could potentially match a given procedure call. 
     Note: OpWam.TRUST_ME is implemented in runWAM() 
     as a pc increment. 
  */


  private void try_me_else(String callLabel, int elseLabelIdx)
  /* Create a choice point which will try the clause 
     starting at label if the clause code following this 
     instruction fails.
     Used for both the WAM ops try_me_else and retry_me_else. 
  */
  { 
    if (picOn)
      if (!callLabel.equals("qqquery_0"))  
        tryNode(callLabel);

    // build a choice point
    ChoicePoint choicePt = new ChoicePoint(args);
    choicePt.label = callLabel;
    choicePt.trailIdx = trail.getLength();
    choicePt.cp = cp;
    choicePt.elseLabel = elseLabelIdx;   // address of next clause
    String elseLabel = getLineLabel(elseLabelIdx);
    choicePt.cutPoint = cutPoint;

    choicePt.callNodeNm = callNodeNm;
    choicePt.tryCount = nodesCount;

    choicePt.envIdx = envIdx;    
        /* this is the 'previous' environment 
           before the one for this clause,
           which will be set up by the call to 
           allocate on the next 
           WAM instruction line */

    if (debugOn)
      ewriteln("  --> Trying " + choicePt.label + 
                      " (" + nodesCount + "); choice recorded"); 

    choices.push(choicePt);   // store the choice point
    pc++;  // start the clause's code

    if (debugOn)
      printChoices();
  }  // end of try_me_else()



  private void backtrack()
  {
    if (debugOn) {
      ewriteln("  --> Backtrack from (" + nodesCount + ")");
      printChoices();
    }

    queryFailed = true;
    if (!choices.isEmpty()) {
      // restore earlier state before trying next clause...
      int choicePos = choices.size()-1;
      ChoicePoint choicePt = choices.pop();

      args = choicePt.args;      // reset the clause head's arguments

      int tIdx = choicePt.trailIdx;   // rollback the trail of bindings
      trail.rollbackTo(tIdx);

      cp = choicePt.cp;          // reset the return address
      pc = choicePt.elseLabel;   // set addr of the next clause to try
      callNodeNm = choicePt.callNodeNm;
      cutPoint = choicePt.cutPoint;

      if (debugOn)
        ewriteln("      Remove C" + choicePos + ": " + 
              choicePt.label + ", try " + getLineLabel(pc));

      envIdx = choicePt.envIdx;
    }
    else {   // should not happen
      trail.rollbackTo(0);
      pc = -1;
    }
  } // end of backtrack()


  // ------------------- CUT Processing ------------------------

  private void get_level(String sv)
  /* v is assigned this clause's choice point as a cut level value
  */
  { 
    Var v = varLookup(sv);
    v.cutLevel = cutPoint;  // cutPoint is set in call()
    pc++;
  } 


  private void cut(String sv)
  /* Discard all the choice points from the current 
     one back to the one stored in v.
  */
  { 
    Var v = varLookup(sv);
    ChoicePoint cpt = v.cutLevel;
    // discard all the choicepoints from the top of
    // the stack down to cpt. 
    if (cpt != null) {
      ChoicePoint currCpt = choices.peek();
      while (!cpt.equals(currCpt)) {
        choices.pop();
        currCpt = choices.peek();
      }
    }
    else
       choices.clear();
    pc++;
  }  // end of cut()


  // ---------------- BUILT-IN PREDICATES --------------------------

  private void callBuiltins(int index)
  /* execute built-in predicates.
     This code is MUCH MORE extensive in Buettcher's system.
  */
  {
    int numArgs = args.size();
    Builtin b = Builtin.fromCode(index);
    if (b == null) {
      System.out.println("Builtin index " + index + " is unknown");
      pc++;
      return;
    }

    if ((b == Builtin.ATOM) && (numArgs == 1)) {
      Var v = args.get(0);
      isAtom(v.deref());
    }
    else if ((b == Builtin.INTEGER) && (numArgs == 1)) {
      Var v = args.get(0);
      integer(v);
    }
    else if ((b == Builtin.VAR) && (numArgs == 1)) {
      Var v = args.get(0);
      var(v);
    }
    else if ((b == Builtin.NONVAR) && (numArgs == 1)) {
      Var v = args.get(0);
      nonvar(v);
    }
    else if ((b == Builtin.WRITELN) && (numArgs == 1)) {
      String valStr = args.get(0).toValue();
      writeln(toText(valStr));
      pc++;
    }
    else if ((b == Builtin.WRITE) && (numArgs == 1)) {
      String valStr = args.get(0).toValue();
      write(toText(valStr));
      pc++;
    }
    else if ((b == Builtin.IS) && (numArgs == 2)) {    // is(Result, v).
      is( args.get(0), args.get(1));
    }
    else if ((b == Builtin.READLN) && (numArgs == 1)) {
      Var v = args.get(0);
      write("> ");
      Var w = new Var(readline());
      if (unify(v,w))
        pc++;
      else
        backtrack();
    }
    else if ((b == Builtin.READ) && (numArgs == 1)) {
      Var v = args.get(0);
      write(">> ");
      Var w = readTerm();
      if (unify(v,w))
        pc++;
      else
        backtrack();
    }
    else if (((b == Builtin.EQ)  ||   // ==
              (b == Builtin.LT)  ||   // <
              (b == Builtin.LE)  ||   // <=
              (b == Builtin.GT)  ||   // >
              (b == Builtin.GE)  ||   // >=
              (b == Builtin.NEQ)) &&
                    (numArgs == 2)) {   // !=
      Var v1 = args.get(0);
      Var v2 = args.get(1);
      comparison(v1, v2, b);
    }
    else if ((b == Builtin.FAIL) && (numArgs == 0)) {
      backtrack();
    }
    else if ((b == Builtin.UNIFY) && (numArgs == 2)) {
      // unify(v1,v2)
      Var v1 = args.get(0);
      Var v2 = args.get(1);
      if (unify(v1,v2))
        pc++;
      else
        backtrack();
    } 
    else if ((b == Builtin.NOT_UNIFY) && (numArgs == 2)) {
      // unify(v1,v2) should fail
      Var v1 = args.get(0);
      Var v2 = args.get(1);
      if (unify(v1,v2))
        backtrack();
      else
        pc++;
    } 
    else if ((b == Builtin.CALL) && (numArgs == 1)) {
      // call(structure)   
      callStruct(args.get(0));
    } 
    else if ((b == Builtin.UNIV) && (numArgs == 3)) {   // =..
      // univ(structure, atom, list of args)
      Var v1 = args.get(0);
      Var v2 = args.get(1);
      Var v3 = args.get(2);
      if (unify_struc2(v1, v2, v3)) {
        v1 = v1.deref();
        pc++;
      }
      else
        backtrack();
    } 
    else {
      System.out.println("Builtin " + b.getName() + 
                     "/" + numArgs + " not implemented");
      pc++;
    }
  } // end of callBuiltins()




  private void isAtom(Var v)
  {
    v = v.deref();
    if ((v.tag == Var.ValType.CON) && isAtom(v.con))
      pc++;
    else
      backtrack();
  } // end of isAtom()




  private void integer(Var v)
  // is v an integer?
  {
    v = v.deref();
    if ((v.tag == Var.ValType.CON) && isInt(v.con))
      pc++;
    else
      backtrack();
  }



  private void var(Var v)
  // is v unbound?
  {
    v = v.deref();
    if (v.tag == Var.ValType.REF)
      pc++;   // yes, it's a variable
    else
      backtrack();
  }



  private void nonvar(Var v)
  // is v instantiated?
  {
    v = v.deref();
    if (v.tag == Var.ValType.REF)
      backtrack();   // no, it's a variable
    else
      pc++;
  }



  private void comparison(Var v1, Var v2, Builtin bop) 
  /* For comparison ops, instead of: ==  <  <=  >  >=  !=
    use: eq, lt, le, gt, ge, neq
  */
  {
    v1 = v1.deref();
    v2 = v2.deref();
    if ((v1.tag == Var.ValType.CON) && (v2.tag == Var.ValType.CON)) {
      int diff = 0;
      try { 
        diff = Integer.parseInt(v1.con) - 
               Integer.parseInt(v2.con);
      }
      catch (Exception e) 
      { System.out.println("Error: comparison must use integers");
        backtrack();
        return;
      }

      switch (bop) {
        case EQ:   // ==
                if (diff == 0)
                  pc++;
                else 
                  backtrack();
                break;

        case LT:   // <
                if (diff < 0) 
                  pc++;
                else 
                  backtrack();
                break;

        case LE:   // <=
                if (diff <= 0)
                  pc++;
                else 
                  backtrack();
                break;

        case GT:   // >
                if (diff > 0)
                  pc++;
                else 
                  backtrack();
                break;

        case GE:   // >=
                if (diff >= 0)
                  pc++;
                else 
                  backtrack();
                break;

        case NEQ:   // !=
                if (diff != 0)
                  pc++;
                else 
                  backtrack();
                break;

        default: 
               System.out.println("Error: Unknown comparison");
               backtrack();
      }
    }
    else
      backtrack();
  } // end of comparison()




  private void is(Var vRes, Var v) 
  // vRes is assigned the result of evaluaing v
  {
    v = v.deref();
    vRes = vRes.deref();

    int result = evalCalc(v);
    if (vRes.tag == Var.ValType.REF) {   // bind vRes to the result
      trail.add(vRes);
      vRes.tag = Var.ValType.CON;
      vRes.con = "" + result;
      pc++;
    }
    else if (vRes.tag == Var.ValType.CON) {
      int val = parseInt(vRes);
      if (val == result)
        pc++;
      else
        backtrack();
    }
    else {
      System.out.println("Error: is/2: result must be a variable or constant");
      backtrack();
    }
  }  // end of is()





  private int evalCalc(Var v)
  /* v can be a structure or an Integer. The structure is
     of arity 2 with functor 
     add, sub/minus, mult/mul, div, mod
  */
  {
    int result = 0;

    if (v.tag == Var.ValType.CON) {
      result = parseInt(v);
    }
    else if (v.tag == Var.ValType.STRUC) {
      String op = v.head.con;
      int[] vals = evalCalcList(v.tail);

      try {   // do the arithmetic
        if (op.equals("add")) 
          result = vals[0] + vals[1];
        else if (op.equals("sub") || op.equals("minus")) 
          result = vals[0] - vals[1];
        else if (op.equals("mult") || op.equals("mul")) 
          result = vals[0] * vals[1];
        else if (op.equals("div")) 
          result = vals[0] / vals[1];
        else if (op.equals("mod")) 
          result = vals[0] % vals[1];
        else
          System.out.println("Error is/2: Unknown operator " + op);
      }
      catch (ArithmeticException e) 
      { System.out.println(" Error is/2: " + e.getMessage());  }
    }
    else 
      System.out.println("Error: bad format for is/2: " + v.tag);

    return result;
  } // end of evalCalc()



  

  private int[] evalCalcList(Var v)
  // v is a list of two arguments used in an is/2 calculation
  {
    int[] args = new int[2];
    args[0] = args[1] = 0;

    v = v.deref();
    if (v.tag == Var.ValType.LIST) {
      args[0] = evalCalc(v.head);
      args[1] = evalCalc(v.tail.deref().head);
    }
    else 
      System.out.println("Error: bad format for is/2 args");

    return args;
  }  // end of evalCalcList()



  private void callStruct(Var v)
  {
    v = v.deref();
    String callLabel = null; 
    int arity = 0;

    if (v.tag == Var.ValType.CON)
      callLabel = v.con;  
    else if (v.tag == Var.ValType.STRUC) {
      callLabel = v.head.con;

      // set up arguments for subsequent call
      Var t = v.tail;  // struct args are in a list
      
      while (t.tag != Var.ValType.CON) {   
        // stop at [] at end of list  ; see put_value
        Var arg = varLookup("A" + arity);
        arg.copyValue(t.head.deref());
        arity++;
        t = t.tail;
      }
    }
    else {   // not a constant or structure
      backtrack();
      return;
    }

    // get code position for the call
    int pos = getCallPos(callLabel, arity);
    if (pos == -1)
      backtrack();
    else
      call( callLabel+"_"+arity, pos);
  }  // end of callStruct()



  private int getCallPos(String callLabel, int arity)
  // get code pos of a user defined predicate or a builtin
  {
    int pos = wc.getLabelPos(callLabel+ "_" + arity);
    if (pos == -1) {
      Builtin b = Builtin.fromName(callLabel);
      if (b != null)
        pos = b.getCode();
    }
    return pos;
  }


  private Var readTerm()
  {
    String line = readline();
    
    if (line == null) {
      writeln("Term string is null");
      return null; 
    }

    line = line.trim();
    if (line.isEmpty()) {
      writeln("term string is empty");
      return null;
    }

    ParseNode termTree = null;
    try {
      BuProlog tc = new BuProlog(line, debugOn);
      termTree = new ParseNode();
      tc.term(termTree, 2);

   //   if (picOn) {
   //     // generate a drawing of the query's parse tree
   //     CallGraph cgTerm = new CallGraph();
   //     termTree.draw(cgTerm, "main");
   //     cgTerm.end();
   //     cgTerm.save("termParse.png");
   //   }
      if (debugOn)
        ewriteln("term tree: " + termTree.toString());
    }
    catch(IOException e)   // usually a parsing error
    {  return null;   }

    return Var.termToVar(termTree);
  }  // end of readTerm()




  // -------------- type checking ----------------------

  // see regexs in lexer.jflex


  private boolean isAtom(String s)
  {  return s.matches("[a-z][a-zA-Z0-9_]*");  }


  private boolean isVar(String s)
  {  return s.matches("[A-Z_][a-zA-Z0-9_]*");  }


  private boolean isString(String s)
  {  return ((s.charAt(0) == '\'') &&
             (s.charAt(s.length()-1) == '\''));
  }


  private String toText(String s)
  {
    if (isString(s))
      return s.substring(1, s.length()-1);
    else
     return s;
  }


  private boolean isInt(String s)
  {  return s.matches("[+-]?[0-9]+");  }


  private int parseInt(Var v)
  {
    int result = 0;
    if (v.tag == Var.ValType.CON) {
      try { 
        result = Integer.parseInt(v.con);
      }
      catch (Exception e) 
      { System.out.println("Error: variable must be an integer");  }
    }
    return result;
  } // end of parseInt()



  // ----------------- I/O utils ---------------------------


  public String readline()
  {
    try {
      return new BufferedReader(
              new InputStreamReader(System.in)).readLine();
    }
    catch (IOException io) {
      return "";
    }
  }  // end of readline()



  private boolean saveString(String fnm, String s) 
  {
    if (s == null)
      return false;

    try {
      FileWriter fw = new FileWriter( new File(fnm));
      fw.write(s);
      fw.close();
      return true;
    } 
    catch (IOException ex) {
      return false;
    } 
  }  // end of saveString()



  public void write(String s)
  {  System.out.print(s);  } 


  public void writeln(String s)
  {  System.out.println(s);  } 


  public void ewrite(String s)
  {  System.err.print(s);  } 


  public void ewriteln(String s)
  {  System.err.println(s);  } 



  private String brief(String s, int len)
  // reduce the length of a string 
  {
    if ((len < 1) || (len > s.length()))
      return s;
    else
     return s.substring(0, len) + "...";
  }



  // --------------- debugging ---------------------------



  private String getLine(int pos)
  // get the WAM statement on this line in the program
  {
    WAMStatement stmt = wc.getStatement(pos);
    return (stmt == null) ? "??" : stmt.toString().trim();
  }



  private String getLineLabel(int pos)
  // get label (pred name) for the WAM statement at this position in the code
  {
    if (pos < -1) {   // may be a built-in
      Builtin b = Builtin.fromCode(pos);
      if (b != null)
        return b.getName();
      else
        return "unknown builtin";
    }

    String line = getLine(pos);
    int idx = line.indexOf(":");    // a line's label is followed by a ":"
    if (idx != -1)
      return line.substring(0, idx);
    else
      return line;   
  }  // end of getLineLabel()



  private String getLineClauseLabel(int pos)
  // get the clause label for this line
  {
    WAMStatement stmt = wc.getStatement(pos);
    return (stmt == null) ? "??" : stmt.getClauseLabel();
  }  // end of getLineClauseLabel()





  private int getArity(String label)
  /* a label may be xxx, xxx_Arity, or xxx_Arity_Choice;
     get the "_Arity" part, or return 0
  */
  {
    String[] nameToks = label.split("_");
    // format is name_arity
    int numArgs = nameToks.length;
    if (numArgs == 1)  
      return 0;
    else 
      return Integer.parseInt(nameToks[1]);
  }  // end of getArity()




  private String stripArity(String label)
  /* a label may be xxx, xxx_Arity, or xxx_Arity_Choice;
     remove the "_Arity" part
  */
  {
    String[] nameToks = label.split("_");
    // format is name_arity
    int numArgs = nameToks.length;
    if (numArgs == 1)  
      return label;
    else if (numArgs == 2)
      return nameToks[0];
    else if (numArgs == 3)
      return (nameToks[0]+ "_" + nameToks[2]);
    else {
      System.out.println("Unknown label format: " + label);
      return label;
    }
  }  // end of stripArity()




  private void printVars(ArrayList<Var> vars, int numVars)
  {
    if (numVars > 0) {
      ewrite("  --> vars: [ ");
      for (int i = 0; i < numVars; i++)  {
        Var v = vars.get(i);
        ewrite(v.toString() + "; ");
      }
      ewriteln("]");
    }
  }  // end of printVars()


  private void printTreeVars(ArrayList<Var> vars, int numVars)
  {
    if (numVars > 0) {
      ewriteln("  --> vars as tree: ");
      for (int i = 0; i < numVars; i++)  {
        Var v = vars.get(i);
        ewriteln(v.toTree(4));
      }
      ewriteln("");
    }
  }  // end of printVars()



  private String getVAs(ArrayList<Var> vars)
  /* return the bindings for PERM vars that refer to args vars
     in the form of a comma-separated string;
     Used in deallocate()
  */
  {
    int numVars = vars.size();
    StringBuilder sb = new StringBuilder();

    for (int i = 0; i < numVars; i++)  {
      Var v = vars.get(i);
      if (v.refersToArg != -1) {
        sb.append(v.toValue() + ", ");
      }
    }
    return sb.toString();
  }  // end of getVAs()



  private String getArgVals(int numArgs)
  /* return the bindings for ARG vars
     in the form of a comma-separated string;
  */
  {
    StringBuilder sb = new StringBuilder("");
    if (numArgs > 0) {
      for (int i = 0; i < numArgs-1; i++)
        sb.append(args.get(i).toValue() + ",");
      sb.append(args.get(numArgs-1).toValue());
    }
    return sb.toString();
  }  // end of getArgVals()





  private void printEnvs()
  // display all of the envList
  {
    int numEnvs = envList.size();
    if (numEnvs > 0) {
      ewriteln("  --> Envs (current env: " + envLabel(envIdx) + "):");
      for (int i = 0; i < numEnvs; i++)  {
        Env env = envList.get(i);
        ewrite("      ");
        ewrite((i == envIdx) ? "* " : "  ");   // mark the current env
        ewriteln("E" + i + ": " + env.clauseLabel + 
             " (" + env.tryCount + "); back to " + 
                                 envLabel(env.ce) );
      }
    }
  }  // end of printEnvs()


  private String envLabel(int idx)
  {  return (idx == -1) ? "query" : "E" + idx;  }



  private void printChoices()
  // display all of the choice points
  {
    int numChoices = choices.size();
    if (numChoices > 0) {
      ewriteln("  --> Choices: ");
      for (int i = 0; i < numChoices; i++)  {
        ChoicePoint cpt = choices.get(i);
        ewrite("      ");
        ewrite((i == numChoices-1) ? "* " : "  "); // mark the current choice
        ewriteln("C" + i + ": " + cpt.label + 
                  " (" + cpt.tryCount + ") in " + 
                  envLabel(cpt.envIdx));
      }
    }
    else
      ewriteln("  --> No Choices");
  }  // end of printChoices()




  // -------------- call graph building -------------------



  private void tryNode(String label)
  {
    tryNodeNm = cg.onCall(callNodeNm, stripArity(label), 
                          "("+ nodesCount + ") try");
  }




  private void callNode(String label)
  {
    // System.out.println("callNode: " + label);
    
    // get parent try node stored in current env
    String tparNm = getTryNode(envIdx);
    if (!Builtin.contains(label))  // user-defined pred
      callNodeNm = cg.onCall(tparNm, stripArity(label), 
                    "("+ nodesCount + ") call\\n args: (" +
                    brief( getArgVals( getArity(label)), 25) + ")");
    else
      callNodeNm = cg.onCall(tparNm, label, 
                    ""+ nodesCount + ") Builtin");

  }  // end of callNode()




  private String getTryNode(int eIdx)
  // get the try node name in the current env
  {
    if (eIdx == -1) // in the top-level query
      return "query";
    else
      return envList.get(eIdx).tryNodeNm;
  } // end of getTryNode()




  // --------------------------------------------------------------

  public static void main(String args[])
  /* reads in the WAM file, then enters a query processing loop
  */ 
  {
    // command line processing of arguments
    if ((args.length < 1) || (args.length > 3)) {
      System.out.println("Usage: java -jar BuWAM.jar [-d] [-p] <WAM file>");
      return;
    }

    String fnm;   // hold the WAM code
    boolean[] flags = new boolean[2];   // { debugOn, picOn }; both false
    if (args.length == 1)
      fnm = args[0];
    else if (args.length == 2) {
      flags = setFlags(new String[]{args[0], ""});
      fnm = args[1];
    }
    else {
      flags = setFlags(new String[]{args[0], args[1]});
      fnm = args[2];
    }

    if (!isWAMFile(fnm)) {
      System.out.println("WAM filename must end in \".wam\"");
      return;
    }

    System.out.println("Welcome to the WAM; type q to quit");
    BuWAM wam = new BuWAM(fnm, flags[0], flags[1]);   // debugOn, picOn

    // query processing loop
    String line;
    do {
      wam.writeln("");
      wam.write("?- ");
      line = wam.readline();
      wam.writeln("");
    } while (wam.runQuery(line));
    wam.writeln("");
  } // end of main()



  private static boolean isWAMFile(String fnm)
  {
    int i = fnm.lastIndexOf('.');
    if (i > 0) {
      String ext = fnm.substring(i+1);
      return ext.equalsIgnoreCase("wam");
    }
    return false;
  }  // end of isWAMFile()



  private static boolean[] setFlags(String[] args)
  {
    boolean debugOn = false;
    boolean picOn = false;

    for (String arg : args) {
       switch(arg.toUpperCase()) {
          case "-D": case "--D":
            debugOn = true; break;

          case "-P": case "--P":
            picOn = true; break;

          case "-DP":   case "-PD":
          case "--DP":  case "--PD":
            debugOn = true; picOn = true; break;

          case "": break;

          default: 
            System.out.println("Unknown flag: " + arg);
            break;
       }
    }
    return new boolean[] { debugOn, picOn };
  }  // end of setFlags()

} // end of class BuWAM

