
// CallGraph.java 
// Andrew Davison, April 2019, ad@fivedots.coe.psu.ac.th

/* Derived from the GraphViz.java code by Laszlo Szathmary
   at https://github.com/jabbalaci/graphviz-java-api

   Assumes that GraphViz (https://www.graphviz.org/) is
   installed.

   A simple API to create a call graph

   Example:
    CallGraph cg = new CallGraph();

    // build call graph nodes

    cg.end());
    System.out.println(cg.getSource());
    cg.save("out.png");

  The node building methods are:
     * onCall()
     * onReturn()
     * comment()

  which can take various arguments.
*/


import java.io.*;
import java.util.*;
import java.nio.file.*;


public class CallGraph
{
  //private static final String GRAPH_VIZ = "Graphviz2.38";
  //     // current stable version for Windows

  // some fill colors
  public static final String LIGHT_BLUE = "\"#9ACEEB\"";
  public static final String LIGHT_ORANGE = "\"#FCD975\"";
  public static final String PINK = "\"#FFCCCC\"";
  public static final String LIME = "\"#E5FFCC\"";
  // more at https://www.rapidtables.com/web/color/RGB_Color.html

  private StringBuilder graphSB;
  private String dotExecStr;

  private int callCount = 1;
  private int codeCount = 1;



  public CallGraph()
  {
    String osName =
             System.getProperty("os.name").replaceAll("\\s", "");
    if (osName.contains("Windows"))
      dotExecStr = winLook("[Gg]raphviz*") + "/bin/dot.exe";
    else if (osName.equals("MacOSX"))
      dotExecStr = "/usr/local/bin/dot";
    else if (osName.equals("Linux"))
      dotExecStr = "/usr/bin/dot";

    if (!new File(dotExecStr).exists()) {
      System.out.println("Could not find path to GraphViz dot");
      dotExecStr = "dot";
    }

    graphSB = new StringBuilder();
    graphSB.append("digraph G {\n");
    graphSB.append("ordering=out;\n");
    graphSB.append("node[shape=Mrecord];\n");
  }  // end of CallGraph()


  private String winLook(String fileRegex)
  // find a path in Windows' program directories that matches fileRegex
  {
    ArrayList<String> fnms = 
             dirLook(System.getenv("ProgramFiles"), fileRegex);
    if (fnms.isEmpty())
      fnms = dirLook(System.getenv("ProgramFiles(x86)"), fileRegex);

    if (fnms.isEmpty())
      return "";    // give up!
    else
      return fnms.get(0);
  } // end of winLook()


  private ArrayList<String> dirLook(String dirName, String fileRegex) 
  // return all the paths in dirName that match fileRegex
  {
    ArrayList<String> fnms = new ArrayList<String>();
    Path dir = Paths.get(dirName);
    PathMatcher pm = FileSystems.getDefault().
                            getPathMatcher("glob:" + fileRegex);
    try (DirectoryStream<Path> dirStream = 
                             Files.newDirectoryStream(dir)) {
      for (Path file : dirStream) {
        if (pm.matches(file.getFileName()))
          fnms.add(file.toAbsolutePath().toString());
      }
    } 
    catch (IOException | DirectoryIteratorException ex) 
    {  System.out.println(ex);  }
    return fnms;
  } // end of dirLook()




  public void end()
  {  graphSB.append("}\n");  }


  public String getSource()
  {  return graphSB.toString(); }


  // public void addLine(String line)
  // {  graphSB.append(line + "\n");  }


  // ----------------------onCall() ---------------------------


  public String onCall(String parentNodeNm, String funcNm)
  {  return onCall(parentNodeNm, funcNm, new String[]{""} );  }


  public String onCall(String parentNodeNm, String funcNm, String data)
  {  return onCall(parentNodeNm, funcNm, new String[]{data} );  }


  public String onCall(String parentNodeNm, String funcNm, String[] data)
  /* create new node for funcNm and its data, then
     link parNode to the new node;
     return the name of the new node
  */
  {
    String nodeNm = addCall(funcNm, data);
    if (nodeNm != null)
      graphSB.append(parentNodeNm + " -> " + nodeNm + ":f0;\n");
    return nodeNm;
  }


  private String addCall(String funcNm, String[] data)
  {
    String nodeNm = funcNm + "_" + callCount;
    callCount++;
    StringBuilder dataCol = new StringBuilder();
    for(String d : data)
      dataCol.append(d + "\\l");     // left justify

    graphSB.append(nodeNm + " [label=\"{<f0> " +
                   funcNm + "|<f1> " + dataCol + "}\"];\n");
    return nodeNm;
  }  // end of addCall()


  // ----------------------comment() ---------------------------


  public String comment(String nodeNm, String msg)
  {  return comment(nodeNm, new String[]{msg}, LIGHT_BLUE );  }

  public String comment(String nodeNm, String[] msgs)
  {  return comment(nodeNm, msgs, LIGHT_BLUE);  }

  public String comment(String nodeNm, String msg, String color)
  {  return comment(nodeNm, new String[]{msg}, color);  }


  public String comment(String nodeNm, String[] msgs, String color)
  /* create box for code messages, and link nodeNm to it 
     return the name of the code box
  */
  {
    String codeNm = nodeNm + "_" + codeCount;
    codeCount++;
    StringBuilder msgsCol = new StringBuilder();
    for(String msg : msgs)
      msgsCol.append(msg + "\\l");     // left justify

    graphSB.append(codeNm + " [label=\"" + msgsCol + 
         "\", shape=box,style=filled,color=" + color + "];\n");
    graphSB.append(nodeNm + " -> " + codeNm + 
             " [color=blue,style=dotted,arrowhead=\"none\"];\n");

    return codeNm;
  }  // end of comment()



  // ----------------------onReturn() ---------------------------


  public String onReturn(String nodeNm, String msg, String parNode, String color)
  {  return onReturn(nodeNm, new String[]{msg}, parNode, color);  }

  public String onReturn(String nodeNm, String msg, String parNode)
  {  return onReturn(nodeNm, new String[]{msg}, parNode, LIGHT_ORANGE);  }


  public String onReturn(String nodeNm, String[] msgs, String parNode)
  {  return onReturn(nodeNm, msgs, parNode, LIGHT_ORANGE);  }  // default color of return


  public String onReturn(String nodeNm, String[] msgs, String parNode, String color)
  /* create shape for return msgs, and link nodeNm to it
     at the same level in the graph. Also invisibly link to
     the parent, so dot doesn't move the node far away.
     return the name of the msgs shape 
  */
  {
    String codeNm = nodeNm + "_" + codeCount;
    codeCount++;
    StringBuilder msgsCol = new StringBuilder();
    for(String msg : msgs)
      msgsCol.append(msg + "\\l");     // left justify

    graphSB.append(codeNm + " [label=\"&#8593; " + msgsCol +  // ^ symbol at start
       "\", shape=house,style=filled,color=" + color + "];\n");
    
    graphSB.append(nodeNm + " -> " + codeNm + 
           " [color=green,penwidth=2,arrowhead=\"none\"];\n");

    graphSB.append("{rank=same;"+nodeNm + ";" + codeNm + "};\n");
    graphSB.append(parNode + " -> " + codeNm + " [style=invis];\n");

    return codeNm;
  }  // end of onReturn()



  // --------------------------------------------------------


  public void setColor(String codeNm, String color)
  /* set the filledcolor attribute for the node
  */
  {  graphSB.append(codeNm + 
            " [style=filled, fillcolor=" + color + "];\n");
  }  



  // ----------------------save() ---------------------------


  public boolean save(String fnm)
  /* Convert graph source into an image using 
     GraphViz dot, and save to the file, fnm
  */
  {
    File dotFile = writeTempSource( getSource() );
    if (dotFile == null)
      return false;

    byte[] img = getImgBytes(dotFile, getFileExt(fnm));
    if (img == null)
      return false;

    try {
      System.out.println("Saving graph image to " + fnm);
      Files.write(new File(fnm).toPath(), img);
      System.out.println("Saved");
      return true;
    }
    catch (IOException e) 
    {  System.out.println("Could not save");
       return false; 
    }
  }  // end of save()



  private String getFileExt(String fnm)
  {
    int idx = fnm.lastIndexOf('.');
    if (idx > 0)
      return fnm.substring(idx+1);
    else
      return "png";
  }



  private File writeTempSource(String str)
  /* Write graph source into a temp file,
     and returns the File object */
  {
    File f = null;
    try {
      f = File.createTempFile("graph_", ".dot.tmp");
      System.out.println("Writing dot data to temp file");
      FileWriter fw = new FileWriter(f);
      fw.write(str);
      fw.close();
    }
    catch (Exception e) {
      System.out.println("Could not wtite dot data to temp file");
    }
    return f;
  }  // end of writeTempSource()




  private byte[] getImgBytes(File dotFile, String imType)
  /* Call the GraphViz dot program, 
     and return the image as a binary array.
  */
  {
    byte[] imgBytes = null;
    try {
      File imFile = File.createTempFile("graph_", "." + imType);
      System.out.println("Processing dot data in GraphViz...");

      if (imType.equals("txt"))
        imType = "plain";

      Runtime rt = Runtime.getRuntime();
      String[] args = { dotExecStr, 
                           "-T" + imType,
                           "-Gdpi=120", // + dpiSizes[dpiIdx],
                           dotFile.getAbsolutePath(), 
                           "-o", imFile.getAbsolutePath() };
      Process p = rt.exec(args);
      p.waitFor();

      // print any GraphViz dot errors
      BufferedReader err = new BufferedReader(new 
                     InputStreamReader(p.getErrorStream()));
      String line = null;
      while ((line = err.readLine()) != null)
        System.out.println(line);

      int retVal = p.exitValue();
      // System.out.println("GraphViz result: " + retVal);
      if (retVal == 0)   // it worked
        imgBytes = Files.readAllBytes(imFile.toPath());
      else
        return null;
    }
    catch (IOException e) {
      System.out.println("GraphViz could not read the dot data");
    }
    catch (java.lang.InterruptedException ie) {
      System.out.println("GraphViz execution was interrupted");
    }
    return imgBytes;
  }  // end of getImgBytes()


  // ------------------- test rig ---------------------

  public static void main(String args[])
  {
    CallGraph cg = new CallGraph();
    cg.onCall("main", "test", "Hello World!");
    cg.end();
    cg.save("test.png");
  }  // end of main()


} // end of CallGraph class
