
/* Lexer.jflex  */
/* Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020  */

/* Compile using jflex.bat
   Test regexs at https://regex101.com

   Multi-line comment regex from:
   https://blog.ostermiller.org/finding-comments-in-source-code-using-regular-expressions/
*/


%%
%{
  private boolean debugOn = false;

  private void echo() 
  { 
    if (debugOn)
      System.err.print(yytext());
  }

  public int position() 
  { return yycolumn; }

  public void error(int pos, String msg)
             throws java.io.IOException
  {
    System.out.println();
    for (int i = 0; i < pos; i++) 
      System.out.print(" ");
    System.out.println("^");
    System.out.println("Lexical error: " + msg);
    throw new java.io.IOException("Lexical error");
    // System.exit(0);
  }  // end of error()

%}

%class    Lexer
%function nextToken
%type	  Token

%ctorarg boolean debugOn

%init{
  this.debugOn = debugOn;
%init}

%unicode
%line             /* line counting on */
%column           /* column counting on */
/* %ignorecase */

%eofval{
  { return new Token( Toks.EOF); }
%eofval}

Atom 	   = [a-z]([:letter:] | [:digit:])*
   /* Note do not add _ to Atom, as this is used by the WAM
      to separate off arity info in the predicate name */

Variable = [A-Z_]([:letter:] | [:digit:])*

Integer  = [-+]?[:digit:]+

Str   = [\'\"][^\'\"]*[\'\"]   
             /* a quoted string; limited to 1 line */


%%
[ \t\b\f\r\n]+	{ echo(); }   /* eat whitespace */

"//"[^\n]*      { echo(); }  /* one-line comment */
"%"[^\n]*       { echo(); }  /* one-line comment */

\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/     
                 { echo(); }  /* multi-line comment */


"("		{ echo(); return new Token( Toks.LPAREN); }
")"		{ echo(); return new Token( Toks.RPAREN); }
"."		{ echo(); return new Token( Toks.PERIOD); }
","		{ echo(); return new Token( Toks.COMMA); }
":-"    { echo(); return new Token( Toks.IF); }

"["		{ echo(); return new Token( Toks.LSQUARE); }
"]"		{ echo(); return new Token( Toks.RSQUARE); }
"|"		{ echo(); return new Token( Toks.BAR); }

"!"		{ echo(); return new Token( Toks.CUT); }


{Atom}	{ echo(); return new Token( Toks.ATOM, yytext()); }

{Variable}	{ echo(); return new Token( Toks.VAR, yytext()); }

{Integer}    { echo(); return new Token( Toks.INTEGER, yytext()); }

{Str}    { echo();  return new Token( Toks.STR,
                      yytext().substring(1, yytext().length()-1));  }


.		{ echo(); error(yycolumn, "illegal char"); }
