
// ParseNode.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020

/* The compiler translates the user's Prolog program into 
   a syntax tree made up of ParseNode objects.

   This tree is translated into WAM code by a call to
   WAMProgram, which recurses over the tree.  
*/


import java.util.*;


public class ParseNode
{
  // public so the data can be accessed without get() methods;
  // think of it as a record

  public ParseType type;
  public String value;
  public ParseNode head, tail;    // links to the subtrees



  // create a structure of type with initial value
  public ParseNode(ParseType t, String v)
  { type = t;
    value = v;
    head = null;
    tail = null;
  } 

  // create a new structure of unknown type
  public ParseNode()
  {  this(ParseType.NO_TYPE, "");  }



  public String toString()
  {
    if (type == ParseType.NO_TYPE)
      return "[no type]";
    else if ((type == ParseType.PREDICATE) ||
             (type == ParseType.CALL) ||
             (type == ParseType.TERM) || 
             (type == ParseType.STRUCTURE) ||
             (type == ParseType.QUERY)) {
      if (tail == null)
        return head.toString();
      else
        return head + "(" + tail + ")";
    }
    else if (type == ParseType.CLAUSES) {
      if (tail == null)
        return "\n" + head;
      else
        return "\n" + head + tail;
    }
    else if (type == ParseType.CLAUSE) {
      if (tail == null)
        return head + ".";
      else
        return head + " :-\n" + tail + ".";
    }
    else if (type == ParseType.BODY) {
      if (tail == null)
        return "  " + head;
      else
        return "  " + head + ",\n" + tail;
    }
    else if ((type == ParseType.ARGS) || 
             (type == ParseType.LIST_ARGS)) {
      if (tail == null)
        return head.toString();
      else
        return head + ", " + tail;
    }
    else if (type == ParseType.LIST) {
      if ((head == null) && (tail == null))
        return "[]";
      else if (tail == null)   // no "|" part
        return "[" + head + "]" ;   // the list args
      else  // args and a | part
        return "[" + head + "|" + tail + "]";
    }
    else if (type == ParseType.ATOM)
      return value;
    else if (type == ParseType.CONSTANT)
      // return "con " + value;
      return value;
    else if (type == ParseType.STR)
      return "'" + value + "'";
    else if (type == ParseType.VARIABLE)
      // return "var " + value;
      return value;
    else
      return "[unknown type]";
  } // end of toString()



  private String spaces(int len)
  // return a string of spaces of the required length
  {
    char[] chs = new char[len];
    Arrays.fill(chs, ' ');
    return new String(chs);
  }



  public void draw(CallGraph cg, String parNode)
  {
    String nodeNm = cg.onCall(parNode, ""+type.getText(), ""+value);
    if (head != null)
      head.draw(cg, nodeNm);
    if (tail != null)
      tail.draw(cg, nodeNm);
  }


} // end of class ParseNode
