
// Var.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, July 2020
// July 24th: added static methods used by read/1

import java.io.*;
import java.util.*;



public class Var
{
  // public so the data can be accessed without get() methods;
  // think of it as a record

  public VarType type = VarType.PERM;  //   PERM, ARG, QUERY

  // a reference; a constant; a list; a structure
  public enum ValType {REF, CON, LIST, STRUC}
  public ValType tag;       

  public int index = 0;    
          // number for the var e.g. V0, V1, etc
          // names the variable when it's not a QUERY var;
          // used for printing a var during debugging 

  public String name = null;     
            // name of var when it's a QUERY variable

  public String con = null;  // content when this var is a CON
  public Var ref;            // content when it is a REF

  public Var head = null;
  public Var tail = null;    
     // used when the var points to a list or structure

  public ChoicePoint cutLevel;  
       // used by the cut and get_level WAM instructions


  // for term reading
  private static TreeMap<String,Integer> varMap = new TreeMap<>();
          // map var names to integers
  private static int varCount = 0;



  // debugging data

  public int refersToArg = -1;
    /* if this var points to a ARG variable then this
       field is set to its index value;
       e.g. if it points to A4, then the value is 4
    */




  // create an unbound variable without a name
  public Var(VarType type, int index)
  { 
    this.type = type;
    this.index = index;
    tag = ValType.REF;
    ref = this;  // an unbound var refers to itself
  }



  public Var(String s) 
  // create a variable with a constant value;
  // used to store text in a var (see readln() in BuWAM.java)
  {
    ref = this;
    tag = ValType.CON; 
    con = s;
  }


  public Var(ValType vt) 
  // create a variable with a given tag;
  {
    ref = this;
    tag = vt; 
  }



  // copy v into a new variable
  public Var(Var v)
  {  
    this.type = v.type;
    this.index = v.index;
    copyValue(v);
  }


  // copy data from v
  public void copyValue(Var v)
  {
    tag = v.tag;
    if (v.type == VarType.ARG) {
      refersToArg = v.index;
      // System.out.println("refers to " + v.index);
    }

    if (tag == ValType.REF)
      ref = v.ref;
    else if (tag == ValType.CON)
      con = v.con;
    else {   // LIST or STRUC
      head = v.head;
      tail = v.tail;
    }
  }  // end of copyValue()



  public Var deref()
  /* if this var points to another var, 
     then return that one dereferenced  */
  {
    if ((tag == ValType.REF) && (ref != this)) {
      Var v = ref;
      while ((v.tag == ValType.REF) && (v.ref != v))
        v = v.ref;
      return v;
    }
    else
      return this;
  } // end of deref()



  public String toString()
  // return the string "varName = VALUE" 
  {
    String varName = name;
    if (varName == null)
      varName = type.getId() + index;    // e.g. "V23"
    return varName + " = " + toValue();
  } // end of toString()




  public String toValue()
  // return VALUE as a string
  {
    if ((tag == ValType.REF) && (ref == this))
      return "_";      // an unbound variable
    else if (tag == ValType.REF)
      return deref().toValue();
    else if (tag == ValType.CON)
      return con;
    else if (tag == ValType.LIST)
      return "[" + argsString() + "]";
    else if (tag == ValType.STRUC) {
      return head.toValue() + "(" + 
             tail.argsString() + ")";
          // the arguments of a structure are stored in a list
    }
    else
      return "";
  } // end of toValue()


  private String argsString()
  // converts a var's list argument into a string
  {
    if (tag == ValType.LIST) {
      String args = head.toValue();
      if (tail == null)   // end of list of args
        return args;
      else {
        Var t = tail.deref();
        if (t.tag == ValType.CON) // end of list is []
          return args;
        else if (t.tag == ValType.LIST)  // more list args
           return args + ", " + t.argsString();
        else if (t.tag == ValType.REF)  // a variable
           return args + ":" + t.toValue();   // GraphViz does not like |
        else // a structure (should not be possible)
           return args + ", " + t.toValue();
      }
    }
    return "";
  } // end of argsString()



  public String toTree(int n)
  // return var as a textual, indented tree
  {
   String varRef = "";
   if (tag == ValType.REF) {
    if (ref == this)
      varRef = " refers to self";
    else
      varRef = " refers to:\n" +  ref.toTree(n+2);
   }

    return  
      ( spaces(n) + type.getId() + index +    
      ((name != null) ? ("(" + name + ")") :"" ) +
      ": " + tag + 
      ((refersToArg != -1) ? 
                ("  ra == " + refersToArg + "; ") : "") +
      ((con != null) ? (" = " + con + "; ") : "" ) +  
       varRef +
      ((head != null) ? ("\n" + spaces(n) + "  head:\n" +
                               head.toTree(n+2)) : "") +
      ((tail != null) ? ("\n" + spaces(n) + "  tail:\n" +
                               tail.toTree(n+2)) : "")
      );
  } // end of toTree()



  private String spaces(int len)
  {
    StringBuffer sb = new StringBuffer(len);
    for (int i = 0; i < len; i++)
       sb.append(" ");
    return sb.toString();
  }


  // ------- convert term ParseNode to a Var --------


  public static Var termToVar(ParseNode t) 
  // conver the term tree into a Var value
  {
    // reset var mapping data structures
    varMap.clear();
    varCount = 0;

    return termGen(t);
  }  // end of termToVar()



  private static Var termGen(ParseNode t) 
  // parse the argument subtree of a term
  {
    Var v = null;   

    if (t.type == ParseType.CONSTANT)
      v = new Var(t.value);
    else if (t.type == ParseType.STR)
      v = new Var("'"+t.value+"'");
    else if (t.type == ParseType.VARIABLE) {
      Integer pos = varMap.get(t.value);
      int id;
      if (pos == null) {
        varMap.put(t.value, varCount);
        id = varCount;
        varCount++;
      }
      else
        id = pos;
      v = makeVar(id);
    }
    else if (t.type == ParseType.STRUCTURE)
      v = structGen(t);
    else if (t.type == ParseType.LIST)
      v = listGen(t);
    else 
      System.out.println("Error: unknown arg type for read/2");

    return v;
  }  // end of termGen()



  private static Var makeVar(int id)
  // store a variable as a var(id) term
  {
    Var vArgs = new Var(ValType.LIST);
    vArgs.head = new Var(""+id);
    vArgs.tail = new Var("[]");

    Var vTerm = new Var(ValType.STRUC);
    vTerm.head = new Var("var");
    vTerm.tail = vArgs;

    return vTerm;
   }  // end of makeVar()



  private static Var structGen(ParseNode t) 
  // parse a STRUCT subtree
  {
    Var v = new Var(ValType.STRUC);
    v.head = new Var(t.head.value);
    v.tail = argsGen(t.tail);
    return v;
  }  // end of structGen()



  private static Var argsGen(ParseNode t) 
  /* parse a sequence of structure (or list) 
     arguments into a Var's list 
  */
  {
    if (t == null)
      return null;
    else {
      Var v = new Var(ValType.LIST);
      v.head = termGen(t.head);  

      if (t.tail == null) // end of arguments: so add [] 
        v.tail = new Var("[]");
      else // otherwise convert the rest of the arguments
        v.tail = argsGen(t.tail);
      return v;
    }
  }  // end of argsGen()



  private static Var listGen(ParseNode t) 
  // parse a LIST subtree
  {
    if ((t.head == null) && (t.tail == null))  // an empty list
      return new Var("[]");

    if (t.tail != null) {   // list ends in a BAR 
      Var barVar = termGen(t.tail);   
      return listArgsGen(t.head, barVar); 
    }
    else
      return argsGen(t.head);
  }  // end of listGen()



  private static Var listArgsGen(ParseNode t, Var barVar) 
    /* convert a sequence of  
       LIST_ARGS nodes into code that generates a list */
  {
    Var v = new Var(ValType.LIST);

    v.head = termGen(t.head);  
                    // process an argument inside a list

    if (t.tail == null) { // end of arguments
      // check for a BAR ending
      if (barVar == null)   // if there's no BAR, add a []
        v.tail = new Var("[]");
      else  // the final variable is the one from the BA
        v.tail = barVar;
    }
    else // otherwise parse the rest of the arguments
      v.tail = listArgsGen(t.tail, barVar); 
   
    return v;
  }  // end of listArgsGen()



} // end of class Var