
// WAMClause.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020

/*  WAMClause passes over the parse tree for a clause (or a
    query) and generates the WAM code, which
    is made up of a series of statements.
    (see WAMStatement.java).

    WAMClause is used by WAMProgram.java from BuPrologr.java, 
    and also by BuWAM.java to translate a user's query.

    The easiest way to understand this code is to generate a
    parse tree diagram (using the -p flag in BuProlog), and 
    then follow the calls on the diagram.
*/

import java.util.*;


public class WAMClause
{

  private ArrayList<WAMStatement> stmts = new ArrayList<>();
    // a clause is a series of WAM statements

  private String label;   // used to identify a clause

  private TreeMap<String,String> varNames = new TreeMap<>();
     // used to rename variables to use the varPrefix 
     // (e.g. YO, Y1, etc)


  private String varPrefix;    
                // "Y" for PERM vars in a clause;
                // "Q" for vars in a query body
     /* the third kind of variables, arguments, 
        are prefixed by "A", and dealt with in
        the code
     */

  private boolean debugOn;



  public WAMClause(ParseNode t, boolean debugOn)
  // parse a CLAUSE or a QUERY
  {
    int numSps = 1;
    this.debugOn = debugOn;
    if (t.type == ParseType.CLAUSE) {
      // this branch is used by BuProlog.java
      if (debugOn) {
        ewriteln("");
        iewriteln(numSps,"clause");
      }
      varPrefix = "Y";
      predGen(t.head, numSps+2);   // a predicate head
      int bodyCalls = 0;
      if (t.tail != null)
        bodyCalls = bodyGen(t.tail, numSps+2);  // an optional body

      if ((varNames.size() > 0) || (bodyCalls > 0)) {
        addToStmts(numSps, 1, new WAMStatement(OpWam.ALLOCATE));
        addToStmts(numSps, new WAMStatement(OpWam.DEALLOCATE));
      }
      addToStmts(numSps, new WAMStatement(OpWam.PROCEED));
    }
    else if (t.type == ParseType.QUERY) {
      // this branch is called from BuWAM.java
      if (debugOn)
        iewriteln(numSps,"query");
      varPrefix = "Q";
      predGen(t.head, numSps+2);   // a dummy predicate name (qqquery)
      bodyGen(t.tail, numSps+2);   // a body
      addToStmts(numSps, new WAMStatement(OpWam.HALT));
    }
    else 
      System.out.println("Error: expected a CLAUSE or QUERY node");
  }  // end of WAMClause()



  private void predGen(ParseNode t, int numSps) 
  // parse the head subtree below PRED into WAM statements
  {
    if (t.type == ParseType.PREDICATE) {
      String name = t.head.value;   // the pred's name
      if (debugOn) {
        iewriteln(numSps,"predGen: " + name);
      }

      int arity = 0;
      int total = 1;
      int count = 1;
      String[] nameToks = name.split("[_/]");
             // format is name/arity_count/total
      if (nameToks.length != 4)
        System.out.println("Name tokenization problem for: \"" + 
                                name + "\"");
      else {
        name = nameToks[0];
        arity = Integer.parseInt(nameToks[1]);
        count = Integer.parseInt(nameToks[2]);
        total = Integer.parseInt(nameToks[3]);
      }
      if (count == 1)
        t.head.value = name + "_" + arity;
      else
        t.head.value = name + "_" + arity + '_' + count;
  
      WAMStatement s;
      if (count < total) {
        if (count > 1)
          s = new WAMStatement(OpWam.RETRY_ME_ELSE, 
                       name + "_" + arity + '_' + (count+1));
        else
          s = new WAMStatement(OpWam.TRY_ME_ELSE, 
                       name + "_" + arity + '_' + (count+1));
      }
      else 
        s = new WAMStatement(OpWam.TRUST_ME);

      s.setLabel(t.head.value);
      label = t.head.value;
      addToStmts(numSps, s);

      // parse the pred's arguments
      if (t.tail != null) { 
        int argCount = 0;
        ParseNode args = t.tail;
        while (args != null) {
          if (args.type == ParseType.ARGS) {
            predArgGen(args.head, argCount, numSps+2);
            argCount++;
          }
          else 
            System.out.println("Error: expected an ARGS node at position " + 
                                (argCount+1) + " for the PRED " +
                                 t.head.value);
          args = args.tail;
        }
      }
    }
    else 
      System.out.println("Error: expected a PREDICATE node");
  }  // end of predGen()



  private void predArgGen(ParseNode t, int argCount, int numSps) 
  // parse the argument subtree of a PRED's ARGS node
  {
    if (debugOn)
      iewriteln(numSps,"predArgGen: " + t.type.getText() + ":" + t.value);

    if (t.type == ParseType.CONSTANT)
      addToStmts(numSps, new WAMStatement(OpWam.GET_CONSTANT, 
                                  t.value, "A" + argCount));
    //
    else if (t.type == ParseType.STR)
      addToStmts(numSps, new WAMStatement(OpWam.GET_CONSTANT, 
                                  "'"+t.value+"'", "A" + argCount));
    //
    else if (t.type == ParseType.VARIABLE) {
      if (isFirstUse(t.value))
        addToStmts(numSps, new WAMStatement(OpWam.GET_VARIABLE, 
                           renameVar(t.value), "A" + argCount));
      else
        addToStmts(numSps, new WAMStatement(OpWam.GET_VALUE, 
                           renameVar(t.value), "A" + argCount));
    }
    //
    else if (t.type == ParseType.STRUCTURE) {
      String v = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.GET_VARIABLE, 
                                  v, "A" + argCount));
      String structVar = structGen(t, numSps+2);
      addToStmts(numSps, new WAMStatement(OpWam.UNIFY_VARIABLE, 
                                  v, structVar));
    }
    //
    else if (t.type == ParseType.LIST) {
      String v = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.GET_VARIABLE, 
                                  v, "A" + argCount));
      String listVar = listGen(t, numSps+2);
      addToStmts(numSps, new WAMStatement(OpWam.UNIFY_VARIABLE, 
                                  v, listVar));
    }
    else 
      System.out.println("Error: not a valid PRED argument node");
  }  // end of predArgGen()




  private int bodyGen(ParseNode t, int numSps) 
  // parse a sequence of BODY nodes
  {
    if (debugOn)
      iewriteln(numSps,"bodyGen");

    int bodyCalls = 0;
    while (t != null) {
      if (t.type == ParseType.BODY) {
        callGen(t.head, numSps+2);
        bodyCalls++;
        t = t.tail;
      }
      else 
        System.out.println("Error: expected a BODY node at position " +
                   (bodyCalls+1));
    }
    return bodyCalls;
  }  // end of bodyGen()




  private void callGen(ParseNode t, int numSps) 
  // parse the CALL subtree
  {
    if (t.type == ParseType.CALL) {
      if (debugOn)
        iewriteln(numSps,"callGen: " + t.head.value);

      if (t.head.value.equals("CUT")) {
        // process a cut
        String v = newVar();
        addToStmts(numSps, 1, new WAMStatement(OpWam.GET_LEVEL, v));
        addToStmts(numSps, new WAMStatement(OpWam.CUT, v));
      }
      else {  // deal with an atom head and args in the tail
        if (t.tail != null) {    // first the arguments
          ParseNode args = t.tail;
          int argCount = 0;
          while (args != null) {
            if (args.type == ParseType.ARGS) {
              callArgGen(args.head, argCount, numSps+2);
              argCount++;
            }
            else 
              System.out.println("Error: expected an ARGS node in position " +
                                  (argCount+1) + " of the CALL to " +
                                   t.head.value);
            args = args.tail;
          }
        }
        // now call the atom
        if (!Builtin.contains(t.head.value)) { // user-defined pred
          String[] nameToks = t.head.value.split("/");
          // format is name/arity
          String name = nameToks[0];
          int arity = Integer.parseInt(nameToks[1]);
          addToStmts(numSps, new WAMStatement(OpWam.CALL, name + "_" + arity));
        }
        else
          addToStmts(numSps, new WAMStatement(OpWam.CALL, t.head.value));
      }
    }
    else 
      System.out.println("Error: expected a CALL node");
  }  // end of callGen()



  private String callArgGen(ParseNode t, int argCount, int numSps) 
  {  return callArgGen(t, argCount, false, numSps);  }


  private String callArgGen(ParseNode t, int argCount, 
                             boolean isInner, int numSps) 
  // parse the argument subtree of a CALL
  // isInner means the argument is inside a list or a struct
  {
    if (debugOn)
      iewriteln(numSps,"callArgGen: " + t.type.getText() + ":" + t.value);

    String lastVar = null;
    if (t.type == ParseType.CONSTANT)
      if (isInner) {
        lastVar = newVar();
        addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                    t.value, lastVar));
      }
      else
        addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                  t.value, "A" + argCount));
    //
    else if (t.type == ParseType.STR)
      if (isInner) {
        lastVar = newVar();
        addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                    "'"+t.value+"'", lastVar));
      }
      else
        addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                "'"+t.value+"'", "A" + argCount));
    //
    else if (t.type == ParseType.VARIABLE) {
      if ( varPrefix.equals("Q") && 
           isFirstUse(t.value) ) // set Q's name
        addToStmts(numSps, new WAMStatement(OpWam.CREATE_VARIABLE, 
                               renameVar(t.value), t.value));
      lastVar = renameVar(t.value);
      if (!isInner)
        addToStmts(numSps, new WAMStatement(OpWam.PUT_VALUE, 
                                    lastVar, "A" + argCount));
    }
    //
    else if (t.type == ParseType.STRUCTURE) {
      lastVar = structGen(t, numSps+2);
      if (!isInner)
        addToStmts(numSps, new WAMStatement(OpWam.PUT_VALUE, 
                                  lastVar, "A" + argCount));
    }
    //
    else if (t.type == ParseType.LIST) {
      lastVar = listGen(t, numSps+2);
      if (!isInner)
        addToStmts(numSps, new WAMStatement(OpWam.PUT_VALUE, 
                                  lastVar, "A" + argCount));
    }
    else 
      System.out.println("Error: unknown arg type for CALL at pos " + 
                         (argCount+1));

    return lastVar;
  }  // end of callArgGen()



  private String structGen(ParseNode t, int numSps) 
  // parse a STRUCT subtree
  {
    if (debugOn)
      iewriteln(numSps,"structGen: " + t.type.getText());

    // process the struct's name
    String headVar = newVar();
    addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                t.head.value, headVar));

    String tailVar;
    if (t.tail != null)    // process the struct's arguments
      tailVar = structArgGen(t.tail, numSps+2);
    else  { // no arguments, so use []
      tailVar = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                           "[]", tailVar));
    }

    // build a structure from the two parts
    String structVar = newVar();
    addToStmts(numSps, new WAMStatement(OpWam.UNIFY_STRUC, 
                                  structVar, 
                                  headVar, tailVar));

    return structVar;
  }  // end of structGen()



  private String structArgGen(ParseNode t, int numSps) 
  /* parse a sequence of structure  
     arguments into code that generates a list 
  */
  {
    if (debugOn)
      iewriteln(numSps,"structArgGen");

    String listVar = null;
    if (t != null) {
      String headVar = callArgGen(t.head, 0, true, numSps+2);  
                       // process an argument inside a struct

      String tailVar;
      if (t.tail == null) { // end of arguments: so add [] 
        tailVar = newVar();
        addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                        "[]", tailVar));
      }
      else // otherwise convert the rest of the arguments
           // into a list using a recursive call
        tailVar = structArgGen(t.tail, numSps+2);

      // build an arguments list from the head and tail
      listVar = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.UNIFY_LIST, 
                           listVar, headVar, tailVar));
    }
    return listVar;
  }  // end of structArgGen()



  private String listGen(ParseNode t, int numSps) 
  // parse a LIST subtree
  {
    if (debugOn)
      iewriteln(numSps,"listGen");

    String listVar;
    if ((t.head == null) && (t.tail == null)) {  // an empty list
      listVar = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                        "[]", listVar));
    }
    else {
      String barVar = null;
      if (t.tail != null)    // list ends in a BAR 
        barVar = callArgGen(t.tail,0, true, numSps+2);   
                     // process BAR argument inside a list
    
      // process the list's arguments (a series of LIST_ARGS nodes)
      // System.out.println("listGen: " + t.type);
      listVar = listArgsGen(t.head, barVar, numSps+2); 
    }
    return listVar;
  }  // end of listGen()




  private String listArgsGen(ParseNode t, String barVar, int numSps) 
    /* convert a sequence of  
       LIST_ARGS nodes into code that generates a list */
  {
    if (debugOn)
      iewriteln(numSps,"listArgsGen");

    String listVar;    
    if (t != null) {  
      String headVar = callArgGen(t.head, 0, true, numSps+2);  
                      // process an argument inside a list

      String tailVar;
      if (t.tail == null) { // end of arguments
        // check for a BAR ending
        if (barVar == null) {   // if there's no BAR, add a []
          tailVar = newVar();
          addToStmts(numSps,  new WAMStatement(OpWam.PUT_CONSTANT, 
                                      "[]", tailVar));
        }
        else  // the final variable is the one from the BAR
          tailVar = barVar;
      }
      else // otherwise parse the rest of the arguments
           // using a recursive call
        tailVar = listArgsGen(t.tail, barVar, numSps+2); 
   
      //  build a list from the head and tail
      listVar = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.UNIFY_LIST, 
                          listVar, headVar, tailVar));

    }
    else { // no arguments, so use []
      listVar = newVar();
      addToStmts(numSps, new WAMStatement(OpWam.PUT_CONSTANT, 
                                  "[]", listVar));
    }
    return listVar;
  }  // end of listArgsGen()



  // ------------------- variable generation ------------


  private String newVar()
  {  return renameVar("");  }   // use a dummy name



  private String renameVar(String varNm)
  {
    boolean hasName = ((varNm.length() > 0) && 
                       (!varNm.equals("_")) );
    if (hasName)
      if (varNames.containsKey(varNm))
        return varNames.get(varNm);
    String newVar = varPrefix + varNames.size();
    if (hasName)
      varNames.put(varNm, newVar);
    else
      varNames.put(newVar, newVar);  // must have a unique key
    return newVar;
  } // end of renameVar()



  private boolean isFirstUse(String varNm)
  {
    if ((varNm.length() > 0) && (!varNm.equals("_")))
      return !varNames.containsKey(varNm);
    return true;
  } 


  // --------------- get WAM code ---------------------



  public String getLabel()
  {  return label;  }


  public ArrayList<WAMStatement> getClause()
  {  return stmts;  }


  public String toString()
  // return the WAM code as a string
  {
    StringBuilder sb = new StringBuilder();
    for (int i = 0; i < stmts.size(); i++)
      sb.append( stmts.get(i).toString() + "\n" );
    return sb.toString();
  } // end of toString()



  // ----------------- debug tools --------------------

  private void ewrite(String s)
  {  System.err.print(s);  } 


  private void ewriteln(String s)
  {  System.err.println(s);  } 


  private void iewriteln(int numSps, String s)
  { 
    if (numSps == 0)
      System.err.printf("\t%c%s\n", '%', s);
    else
      System.err.printf("\t%c%"+numSps+"s%s\n", '%', "", s);
  }



  private void addToStmts(int numSps, WAMStatement stmt)
  {
    if (debugOn) {
      String stmtStr = stmt.toString().trim();
      if (numSps == 0)
        System.err.printf("\t|[ %s ]\n", stmtStr);
      else
        System.err.printf("\t|%"+numSps+"s[ %s ]\n", "", stmtStr);
    }
    stmts.add(stmt);
  }  // end of addToStmts()



  private void addToStmts(int numSps, int pos, WAMStatement stmt)
  {
    if (debugOn) {
      String stmtStr = stmt.toString().trim();
      if (numSps == 0)
        System.err.printf("\t|[ %s ] (%d)\n", stmtStr, pos);
      else
        System.err.printf("\t|%"+numSps+"s[ %s ] (%d)\n", "", stmtStr, pos);
    }
    stmts.add(pos, stmt);
  }  // end of addToStmts()



}  // end of WAMClause class