
// WAMCode.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020

/* WAM code consists of a list of WAM statements
   (see WAMStatement.java). The list is built by reading in
   lines of text from a WAM textfile.

   At run-time, the code for a compiled query is temporarily added
   to this program at the end as an extra "query" clause.

   After the query has been run, it is deleted from the code.

   WAMCode is called by the WAM, while WAMProgram is called by
   BuProlog. Both of them use WAMStatement.
*/

import java.awt.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;


public class WAMCode
{

  private ArrayList<WAMStatement> stmts = new ArrayList<>();
  private TreeMap<String, Integer> labels = new TreeMap<>();
     // stores label text --> line number mappings in sorted key order

  // the label and position of a loaded query clause
  private String queryLabel = null; 
  private int queryLabelPos = -1; 

  Pattern toksPat;

  private boolean debugOn;



  public WAMCode(String fnm, boolean debugOn)
  {
    this.debugOn = debugOn;

    toksPat = Pattern.compile("\'[^\']*\'|\\S+");
                // words and 'xxxx' strings

    readProgram(fnm);   // get the WAM code
    if (stmts.size() == 0)
      writeln("Could not load " + fnm);
    else {
      setLabels();
      setJumps(0, stmts.size());
      writeln("Loaded " + fnm);
      if (debugOn) {
        ewriteln("\n----------------- WAM code ----------------");
        ewriteln(toString());
        ewriteln("--------------------------------------------\n");
        ewrite("Labels=Jumps: ");
        ewriteln(labels.toString() + "\n");
      }
    }
  }  // end of WAMCode()




  private void setLabels()
  // map the code's labels to their positions
  {
    labels.clear();
    String label;
    for (int i = 0; i < stmts.size(); i++) {
      WAMStatement stmt = stmts.get(i);
      label = stmt.getLabel();
      if (label.length() > 0)
        labels.put(label, i);
    }
  } // end of setLabels()



  public int getLabelPos(String label)
  // find a label's code positions
  {
     Integer pos = labels.get(label);
     if (pos == null) {
       // System.out.println("Error: no code found with label \"" + 
       //                           label + "\"");
       return -1;
     }
     else 
       return pos.intValue();
  } // end of getLabelPos()



  private void setJumps(int start, int end)
  /* Store the jumps for the code between the lines 
     at start to end-1 inclusive

     It converts label names in CALL, TRY_ME_ELSE, 
     and RETRY_ME_ELSE stmts
     to integers. Builtin predicate names (e.g. writeln) 
     are transformed into negative numbers.
  */
  {
    String label;
    for (int i = start; i < end; i++) {
      WAMStatement stmt = stmts.get(i);
      OpWam op = stmt.getOp();
      if ((op == OpWam.CALL) ||
          (op == OpWam.TRY_ME_ELSE) ||
          (op == OpWam.RETRY_ME_ELSE)) {
        label = stmt.getArg(0);
        stmt.setJump(-1);   // -1 denotes an error
        if (labels.containsKey(label)) 
          // label is a user-defined predicate
          stmt.setJump((labels.get(label)).intValue());
        else { 
          // label might be a built-in predicate
          Builtin b = Builtin.fromName(label);
          if (b != null)
            stmt.setJump(b.getCode());
          else
            System.out.println("Error: \"" + stripArity(label) + 
                               "\" predicate undefined");
        }
      }
    }
  }  // end of setJumps()



  private String stripArity(String label)
  //  remove the arity text from the label: xxxx_N becomes xxxx
  {
    String[] nameToks = label.split("_");
    return nameToks[0];
  }




  public WAMStatement getStatement(int i)
  {  
    if ((i < 0) || (i >= stmts.size()))
      return null;
    else
      return stmts.get(i);
  }


  public void addQuery(WAMClause query)
  {
    queryLabel = query.getLabel();
    queryLabelPos = stmts.size();
    stmts.addAll(query.getClause());
    labels.put(queryLabel, queryLabelPos);

    setJumps(queryLabelPos, stmts.size());  
            // update the query's labels for jumping

   if (debugOn) {
     ewriteln("\n----------------- WAM code ----------------");
     ewriteln(toNumString(queryLabelPos, stmts.size()));
     ewriteln("--------------------------------------------\n");
     ewrite("Labels=Jumps: ");
     ewriteln(labels.toString() + "\n");
   }
  } // end of addQuery()



  public int getQueryPos()
  {  return queryLabelPos;  }



  public void deleteQuery()
  {
    if (queryLabelPos != -1) {      
      int end = stmts.size();
      stmts.subList(queryLabelPos, end).clear();   // removes the range
	  
      labels.remove(queryLabel);
      queryLabel = null;
      queryLabelPos = -1;
    }
  } // end of deleteQuery()



  public String toString()
  {  return toNumString(0, stmts.size());  }




  public String toNumString(int start, int end)
  // add a 4-digit line number to the front of each statement
  {
    StringBuilder sb = new StringBuilder();
    for (int i = start; i < end; i++) {
      sb.append( "(" + String.format("%04d", i) + 
                      ")  " + stmts.get(i).toString() );
      if (i < stmts.size()-1) 
        sb.append("\n");
    }
    return sb.toString();
  } // end of toNumString()




  // ------------------ load WAM file ------------------



  private void readProgram(String fnm)
  {
    BufferedReader br;
    try {
      br = new BufferedReader(new FileReader(fnm));
      WAMStatement stmt;
      String line;
      String clauseLabel = "";

      do {
        line = br.readLine();
        if (line != null) {
          line = line.trim();
          if (line.length() == 0)
            continue;
          if ((line.charAt(0) == ';') || 
              (line.charAt(0) == '#') || 
              (line.charAt(0) == '%'))   // ignore comment lines
            continue;

          String label = "";
          int j = line.indexOf(":::");
          if (j > 0) {
            label = line.substring(0, j).trim();
            line = line.substring(j + 3).trim();
          }
          if (!label.equals(""))   // found a new label
            clauseLabel = new String(label);

          String[] toks = getToks(line);
          if (toks.length < 1) {
            writeln("Missing operation on line with label \"" + 
                    label + "\"");
            continue;
          }

          OpWam op = OpWam.fromString(toks[0].trim());
          if (toks.length == 1)
            stmt = new WAMStatement(op);
          else if (toks.length == 2)
            stmt = new WAMStatement(op, toks[1]);
          else if (toks.length == 3)
            stmt = new WAMStatement(op, toks[1], toks[2]);
          else if (toks.length == 4)
            stmt = new WAMStatement(op, toks[1], toks[2], toks[3]);
          else {
            writeln("Too many tokens on line \"" + line + 
                    "\"; using first four");
            stmt = new WAMStatement(op, toks[1], toks[2], toks[3]);
          }
          stmt.setLabel(label);
          stmt.setClauseLabel(clauseLabel);
          stmts.add(stmt);
        }
      } while (line != null);
      br.close();
    }
    catch (IOException e) {
      // writeln(e.toString());
    }
  } // end of readProgram()


  private String[] getToks(String s)
  {
    ArrayList<String> matches = new ArrayList<String>();
    Matcher m = toksPat.matcher(s);
    while (m.find()) 
      matches.add(m.group());
    return matches.toArray(new String[0]);
  }  // end of getToks()


  public void write(String s)
  {  System.out.print(s);  } 


  public void writeln(String s)
  {  System.out.println(s);  } 


  public void ewrite(String s)
  {  System.err.print(s);  } 


  public void ewriteln(String s)
  {  System.err.println(s);  } 



} // end of class WAMCode
