
// WAMStatement.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, June 2020

/* WAMStatement represents a single line of WAM code.

   The format of a statement:
       [label:::] opName operand1 [ operand2 [operand3] ]

   A WAM operation may have upto three operands.

   The label only appears on the first line of a clause,
   and is used when CALL ops are executed to set the
   program counter.

  A statement also stores a clause label, which is means that
  every line has information about the clause it is part of.

  A statement also stores a jump int, which is used to indicate
  what progam line number a CALL will jump to.
 
*/

import java.util.*;


public class WAMStatement
{

  private String label = "";    
        // label which only appears on the first line of a clause
  private String clauseLabel = "query";    
        // the clause's label; default is "query"

  private OpWam wop;                  // WAM opcode
  private ArrayList<String> args;     // operands list

  private int jump = -1;     
     /* for jumping to another clause;
        used by the CALL, TRY_ME_ELSE, and RETRY_ME_ELSE opcode
     */


  // creates a statement with up to three arguments
  public WAMStatement(OpWam op, 
                   String arg1, String arg2, String arg3)
  {
    wop = op;
    args = new ArrayList<>();
    if (arg1 != null)
      args.add(arg1);
    if (arg2 != null)
      args.add(arg2);
    if (arg3 != null)
      args.add(arg3);
  } // end of WAMStatement()


  // creates a new statement with no arguments
  public WAMStatement(OpWam op)
  {  this(op, null, null, null);  }


  // creates a new statement with one argument
  public WAMStatement(OpWam op, String arg1)
  {  this(op, arg1, null, null);  }


  // creates a new statement with two arguments
  public WAMStatement(OpWam op, String arg1, String arg2)
  {  this(op, arg1, arg2, null);  }


  public OpWam getOp()
  {  return wop; } 


  public String getArg(int i)
  {
    if ((i < 0) || (i >= args.size())) {
      System.out.println("No WAM statement argument " + i + 
                         " for " + wop.getName());
      return "";
    }
    else
      return args.get(i);
  }  // end of getArg()



  public void setLabel(String l)
  {  label = l.trim();  } 

  public String getLabel()
  {  return label; } 


  public void setClauseLabel(String l)
  {  clauseLabel = l.trim();  } 

  public String getClauseLabel()
  {  return clauseLabel; } 


  public void setJump(int addr)
  {  jump = addr; } 

  public int getJump()
  {  return jump;  } 



  public String toString()
  /* return the statement as a string: 
         "[label:::] op_code op1 op2"

     clauseLabel and jump are not printed 
  */
  {
    StringBuilder sb = new StringBuilder();
    if (label.length() > 0)
      sb.append( String.format("%16s", label+"::: "));
    else
      sb.append("                ");

    sb.append( wop.getName() );
    for (int i = 0; i < args.size(); i++) {
      String a = (String) args.get(i);
    // if (a.indexOf(' ') < 0)
        sb.append(" " + a);
    //  else
    //    sb.append( " '" + a + "'");
    }
    return sb.toString();
  } // end of toString()



} // end of class WAMStatement
